from __future__ import annotations

import logging
import math
import pickle

import psutil

from libtbx import Auto

import dials.algorithms.integration
from dials.algorithms.integration.processor import NullTask, execute_parallel_task
from dials.array_family import flex
from dials.util import tabulate
from dials.util.mp import multi_node_parallel_map

# Need this import first because loads extension that parallel_integrator_ext
# relies on - it assumes the binding for EmpiricalProfileModeller exists
import dials.algorithms.profile_model.modeller  # noqa: F401 # isort: split

from dials.constants import EPS, FULL_PARTIALITY
from dials_algorithms_integration_parallel_integrator_ext import (
    GaussianRSIntensityCalculator,
    GaussianRSMaskCalculator,
    GaussianRSMultiCrystalMaskCalculator,
    GaussianRSMultiCrystalReferenceProfileData,
    GaussianRSReferenceCalculator,
    GaussianRSReferenceProfileData,
    GLMBackgroundCalculator,
    Logger,
    MultiThreadedIntegrator,
    MultiThreadedReferenceProfiler,
    ReferenceProfileData,
    SimpleBackgroundCalculator,
    SimpleBlockList,
    SimpleReflectionManager,
)

__all__ = [
    "BackgroundCalculatorFactory",
    "GLMBackgroundCalculator",
    "GaussianRSIntensityCalculator",
    "GaussianRSMaskCalculator",
    "GaussianRSMultiCrystalMaskCalculator",
    "GaussianRSMultiCrystalReferenceProfileData",
    "GaussianRSReferenceCalculator",
    "GaussianRSReferenceProfileData",
    "IntegrationJob",
    "IntegrationManager",
    "IntegratorProcessor",
    "IntensityCalculatorFactory",
    "Logger",
    "MaskCalculatorFactory",
    "MultiThreadedIntegrator",
    "MultiThreadedReferenceProfiler",
    "ReferenceCalculatorFactory",
    "ReferenceCalculatorJob",
    "ReferenceCalculatorManager",
    "ReferenceCalculatorProcessor",
    "ReferenceProfileData",
    "SimpleBackgroundCalculator",
    "SimpleBlockList",
    "SimpleReflectionManager",
]

logger = logging.getLogger(__name__)


class MaskCalculatorFactory:
    """
    A factory function to return a mask calculator object
    """

    @staticmethod
    def create(experiments, params=None):
        """
        Select the mask calculator
        """
        from dials.algorithms.profile_model.gaussian_rs.algorithm import (
            GaussianRSMaskCalculatorFactory,
        )

        # Get the parameters
        if params is None:
            from dials.command_line.integrate import phil_scope

            params = phil_scope.extract()

        # Select the factory function
        selection = params.profile.algorithm
        if selection == "gaussian_rs":
            algorithm = GaussianRSMaskCalculatorFactory.create(experiments)
        else:
            raise RuntimeError("Unknown profile model algorithm")

        # Create the mask algorithm
        return algorithm


class BackgroundCalculatorFactory:
    """
    A factory function to return a background calculator object
    """

    @staticmethod
    def create(experiments, params=None):
        """
        Select the background calculator
        """
        from dials.algorithms.background.glm.algorithm import (
            GLMBackgroundCalculatorFactory,
        )
        from dials.algorithms.background.gmodel.algorithm import (
            GModelBackgroundCalculatorFactory,
        )
        from dials.algorithms.background.simple.algorithm import (
            SimpleBackgroundCalculatorFactory,
        )

        # Get the parameters
        if params is None:
            from dials.command_line.integrate import phil_scope

            params = phil_scope.extract()

        # Select the factory function
        selection = params.integration.background.algorithm
        if selection == "simple":

            # Get parameters
            params = params.integration.background.simple

            # Create some keyword parameters
            kwargs = {
                "model": params.model.algorithm,
                "outlier": params.outlier.algorithm,
                "min_pixels": params.min_pixels,
            }

            # Create all the keyword parameters
            if params.outlier.algorithm == "null":
                pass
            elif params.outlier.algorithm == "truncated":
                kwargs["lower"] = params.outlier.truncated.lower
                kwargs["upper"] = params.outlier.truncated.upper
            elif params.outlier.algorithm == "nsigma":
                kwargs["lower"] = params.outlier.nsigma.lower
                kwargs["upper"] = params.outlier.nsigma.upper
            elif params.outlier.algorithm == "normal":
                kwargs["min_pixels"] = params.outlier.normal.min_pixels
            elif params.outlier.algorithm == "plane":
                kwargs["fraction"] = params.outlier.plane.fraction
                kwargs["n_sigma"] = params.outlier.plane.n_sigma
            elif params.outlier.algorithm == "tukey":
                kwargs["lower"] = params.outlier.tukey.lower
                kwargs["upper"] = params.outlier.tukey.upper

            # Create the algorithm
            algorithm = SimpleBackgroundCalculatorFactory.create(experiments, **kwargs)

        elif selection == "glm":

            # Get the parameters
            params = params.integration.background.glm

            # Create the algorithm
            algorithm = GLMBackgroundCalculatorFactory.create(
                experiments,
                model=params.model.algorithm,
                tuning_constant=params.robust.tuning_constant,
                min_pixels=params.min_pixels,
            )

        elif selection == "gmodel":

            # Get the parameters
            params = params.integration.background.gmodel

            # Create the algorithm
            algorithm = GModelBackgroundCalculatorFactory.create(
                experiments,
                model=params.model,
                robust=params.robust.algorithm,
                min_pixels=params.min_pixels,
            )

        else:
            raise RuntimeError("Unknown background algorithm")

        # Return the background calculator
        return algorithm


class IntensityCalculatorFactory:
    """
    A factory function to return an intensity calculator object
    """

    @staticmethod
    def create(experiments, reference_profiles, params=None):
        """
        Select the intensity calculator
        """
        from dials.algorithms.profile_model.gaussian_rs.algorithm import (
            GaussianRSIntensityCalculatorFactory,
        )

        # Get the parameters
        if params is None:
            from dials.command_line.integrate import phil_scope

            params = phil_scope.extract()

        # Select the factory function
        selection = params.profile.algorithm
        if selection == "gaussian_rs":

            # Get the parameters
            params = params.profile.gaussian_rs.fitting

            # Check for detector space
            if params.fit_method == "reciprocal_space":
                detector_space = False
            elif params.fit_method == "detector_space":
                detector_space = True
            else:
                raise RuntimeError(f"Unknown fit method: {params.fit_method}")

            # Create the algorithm
            algorithm = GaussianRSIntensityCalculatorFactory.create(
                reference_profiles,
                detector_space=detector_space,
                deconvolution=params.detector_space.deconvolution,
            )

        else:
            raise RuntimeError("Unknown profile model algorithm")

        # Return the algorithm
        return algorithm


class ReferenceCalculatorFactory:
    """
    A factory function to return an reference calculator object
    """

    @staticmethod
    def create(experiments, params=None):
        """
        Select the reference calculator
        """
        from dials.algorithms.profile_model.gaussian_rs.algorithm import (
            GaussianRSReferenceCalculatorFactory,
        )

        # Get the parameters
        if params is None:
            from dials.command_line.integrate import phil_scope

            params = phil_scope.extract()

        # Select the factory function
        selection = params.profile.algorithm
        if selection == "gaussian_rs":

            # Get the parameters
            params = params.profile.gaussian_rs.fitting

            # Create the algorithm
            algorithm = GaussianRSReferenceCalculatorFactory.create(
                experiments,
                grid_size=params.grid_size,
                scan_step=params.scan_step,
                grid_method=params.grid_method,
            )

        else:
            raise RuntimeError("Unknown profile model algorithm")

        # Return the algorithm
        return algorithm


def _assert_enough_memory(required_memory, max_memory_usage):
    """
    Check there is enough memory available or fail

    :param required_memory: The required number of bytes
    :param max_memory_usage: The maximum memory usage allowed
    """
    total_memory = psutil.virtual_memory().total
    assert max_memory_usage > 0.0, "maximum memory usage must be > 0"
    assert max_memory_usage <= 1.0, "maximum memory usage must be <= 1"
    limit_memory = total_memory * max_memory_usage
    if required_memory > limit_memory:
        raise RuntimeError(
            f"""
    There was a problem allocating memory for image data. Possible solutions
    include increasing the percentage of memory allowed for shoeboxes or
    decreasing the block size. This could also be caused by a highly mosaic
    crystal model - is your crystal really this mosaic?
      Total system memory: {total_memory / 1000000000.0:.1f} GB
      Limit image memory: {limit_memory / 1000000000.0:.1f} GB
      Required image memory: {required_memory / 1000000000.0:.1f} GB
    """
        )
    else:
        logger.info("Allocating %.1f MB memory", required_memory / 1e6)


class IntegrationJob:
    """
    A class to represent an integration job
    """

    def __init__(self, index, job, experiments, reflections, reference, params=None):
        """
        Initialise the task.

        :param index: The index of the processing job
        :param experiments: The list of experiments
        :param reflections: The list of reflections
        :param params: The processing parameters
        :param job: The frames to integrate
        :param flatten: Flatten the shoeboxes
        :param executor: The executor class
        """

        # Get the parameters
        if params is None:
            from dials.command_line.integrate import phil_scope

            params = phil_scope.extract()

        assert len(reflections) > 0, "Zero reflections given"
        assert (
            params.integration.block.max_memory_usage > 0.0
        ), "Max memory % must be > 0"
        assert (
            params.integration.block.max_memory_usage <= 1.0
        ), "Max memory % must be < 1"
        self.index = index
        self.job = job
        self.experiments = experiments
        self.reflections = reflections
        self.reference = reference
        self.params = params

    def __call__(self):
        """
        Do the processing.

        :return: The processed data
        """
        from dials.algorithms.integration.processor import job

        # Set the global process ID
        job.index = self.index

        # Check all reflections have same imageset and get it
        imageset = self.experiments[0].imageset
        if not all(e.imageset == imageset for e in self.experiments):
            raise RuntimeError("Task can only handle 1 imageset")

        # Get the sub imageset
        frame0, frame1 = self.job
        try:
            frame10, frame11 = imageset.get_array_range()
        except Exception:
            frame10, frame11 = (0, len(imageset))
        try:
            assert frame0 < frame1
            assert frame10 < frame11
            assert frame0 >= frame10
            assert frame1 <= frame11
            index0 = frame0 - frame10
            index1 = index0 + (frame1 - frame0)
            assert index0 < index1
            assert index0 >= 0
            assert index1 <= len(imageset)
            imageset = imageset[index0:index1]
        except Exception:
            raise RuntimeError("Programmer Error: bad array range")

        # Check the memory requirements
        _assert_enough_memory(
            self.compute_required_memory(imageset),
            self.params.integration.block.max_memory_usage,
        )

        # Integrate
        self.integrate(imageset)

        # Write some debug files
        self.write_debug_files()

        # Return the result
        return dials.algorithms.integration.Result(
            index=self.index,
            reflections=self.reflections,
            data=None,
            read_time=0,
            extract_time=0,
            process_time=0,
            total_time=0,
        )

    def compute_required_memory(self, imageset):
        """
        Compute the required memory
        """
        return MultiThreadedIntegrator.compute_required_memory(
            imageset, self.params.integration.block.size
        )

    def integrate(self, imageset):
        """
        Integrate the reflections
        """
        from dials.algorithms.integration.integrator import frame_hist

        # Compute the partiality
        self.reflections.compute_partiality(self.experiments)

        # Get some info
        full_value = FULL_PARTIALITY - EPS
        fully_recorded = self.reflections["partiality"] > full_value
        npart = fully_recorded.count(False)
        nfull = fully_recorded.count(True)
        select_ice = self.reflections.get_flags(self.reflections.flags.in_powder_ring)
        select_int = ~self.reflections.get_flags(self.reflections.flags.dont_integrate)
        nice = select_ice.count(True)
        nint = select_int.count(True)
        ntot = len(self.reflections)
        frame0, frame1 = imageset.get_scan().get_array_range()

        # Write some output
        logger.info(" Beginning integration job %d", self.index)
        logger.info("")
        logger.info(" Frames: %d -> %d", frame0, frame1)
        logger.info("")
        logger.info(" Number of reflections")
        logger.info("  Partial:     %d", npart)
        logger.info("  Full:        %d", nfull)
        logger.info("  In ice ring: %d", nice)
        logger.info("  Integrate:   %d", nint)
        logger.info("  Total:       %d", ntot)
        logger.info("")

        # Print a histogram of reflections on frames
        if frame1 - frame0 > 1:
            logger.info(
                " The following histogram shows the number of reflections predicted"
            )
            logger.info(" to have all or part of their intensity on each frame.")
            logger.info("")
            logger.info(
                frame_hist(
                    self.reflections["bbox"].select(select_int), prefix=" ", symbol="*"
                )
            )
            logger.info("")

        # Construct the mask algorithm
        compute_mask = MaskCalculatorFactory.create(self.experiments, self.params)

        # Construct the background algorithm
        compute_background = BackgroundCalculatorFactory.create(
            self.experiments, self.params
        )

        # Construct the intensity algorithm
        compute_intensity = IntensityCalculatorFactory.create(
            self.experiments, self.reference, self.params
        )

        # Call the multi threaded integrator
        integrator = MultiThreadedIntegrator(
            reflections=self.reflections,
            imageset=imageset,
            compute_mask=compute_mask,
            compute_background=compute_background,
            compute_intensity=compute_intensity,
            logger=Logger(logger),
            nthreads=self.params.integration.mp.nproc,
            buffer_size=self.params.integration.block.size,
            use_dynamic_mask=self.params.integration.use_dynamic_mask,
            debug=self.params.integration.debug.output,
        )

        # Assign the reflections
        self.reflections = integrator.reflections()

    def write_debug_files(self):
        """
        Write some debug output
        """

        # Optionally save the shoeboxes
        debug = self.params.integration.debug
        if debug.output and debug.separate_files:
            output = self.reflections
            if debug.select is not None:
                output = output.select(debug.select(output))
            if debug.split_experiments:
                output = output.split_by_experiment_id()
                for table in output:
                    i = table["id"][0]
                    table.as_file("shoeboxes_%d_%d.refl" % (self.index, i))
            else:
                output.as_file("shoeboxes_%d.refl" % self.index)

        # Delete the shoeboxes
        if debug.separate_files or not debug.output:
            del self.reflections["shoebox"]


class IntegrationManager:
    """
    A class to manage processing book-keeping
    """

    def __init__(self, experiments, reflections, reference, params):
        """
        Initialise the manager.

        :param experiments: The list of experiments
        :param reflections: The list of reflections
        :param reference: The reference profiles
        :param params: The phil parameters
        """

        # Save some data
        self.experiments = experiments
        self.reflections = reflections
        self.reference = reference

        # Save some parameters
        self.params = params

        # Set the finalized flag to False
        self.finalized = False

        # Initialise the timing information
        # self.time = TimingInfo()

        self.initialize()

    def initialize(self):
        """
        Initialise the processing
        """
        # Ensure the reflections contain bounding boxes
        assert "bbox" in self.reflections, "Reflections have no bbox"

        # Compute the block size and jobs
        self.compute_blocks()
        self.compute_jobs()
        self.reflections = split_partials_over_boundaries(
            self.reflections, self.params.integration.block.size
        )

        # Create the reflection manager
        self.manager = SimpleReflectionManager(
            self.blocks, self.reflections, self.params.integration.mp.njobs
        )

    def task(self, index):
        """
        Get a task.
        """
        frames = self.manager.job(index)
        experiments = self.experiments
        reference = self.reference
        reflections = self.manager.split(index)
        if len(reflections) == 0:
            logger.warning("No reflections in job %d ***", index)
            task = NullTask(index=index, reflections=reflections)
        else:
            task = IntegrationJob(
                index=index,
                job=frames,
                experiments=experiments,
                reflections=reflections,
                reference=reference,
                params=self.params,
            )
        return task

    def tasks(self):
        """
        Iterate through the tasks.
        """
        for i in range(len(self)):
            yield self.task(i)

    def accumulate(self, result):
        """Accumulate the results."""
        self.manager.accumulate(result.index, result.reflections)
        # self.time.read += result.read_time
        # self.time.extract += result.extract_time
        # self.time.process += result.process_time
        # self.time.total += result.total_time

    def finalize(self):
        """
        Finalize the processing and finish.
        """
        # Check manager is finished
        assert self.manager.finished(), "Manager is not finished"

        self.finalized = True

    def result(self):
        """
        Return the result.

        :return: The result
        """
        assert self.finalized, "Manager is not finalized"
        return self.manager.data()

    def finished(self):
        """
        Return if all tasks have finished.

        :return: True/False all tasks have finished
        """
        return self.finalized and self.manager.finished()

    def __len__(self):
        """
        Return the number of tasks.

        :return: the number of tasks
        """
        return len(self.manager)

    def compute_max_block_size(self):
        """
        Compute the required memory
        """
        total_memory = psutil.virtual_memory().available
        max_memory_usage = self.params.integration.block.max_memory_usage
        assert max_memory_usage > 0.0, "maximum memory usage must be > 0"
        assert max_memory_usage <= 1.0, "maximum memory usage must be <= 1"
        limit_memory = int(math.floor(total_memory * max_memory_usage))
        return MultiThreadedIntegrator.compute_max_block_size(
            self.experiments[0].imageset, max_memory_usage=limit_memory
        )

    def compute_blocks(self):
        """
        Compute the processing block size.
        """
        block = self.params.integration.block
        max_block_size = self.compute_max_block_size()
        if block.size in [Auto, "auto", "Auto"]:
            assert block.threshold > 0, "Threshold must be > 0"
            assert block.threshold <= 1.0, "Threshold must be < 1"
            nframes = sorted([b[5] - b[4] for b in self.reflections["bbox"]])
            cutoff = int(block.threshold * len(nframes))
            block_size = nframes[cutoff]
            if block_size > max_block_size:
                logger.warning(
                    "Computed block size (%s) > maximum block size (%s).",
                    block_size,
                    max_block_size,
                )
                logger.warning(
                    "Setting block size to maximum; some reflections may be partial"
                )
                block_size = max_block_size
        else:
            scan = self.experiments[0].scan
            if block.units == "radians":
                phi0, dphi = scan.get_oscillation(deg=False)
                block_size = int(math.ceil(block.size / dphi))
            elif block.units == "degrees":
                phi0, dphi = scan.get_oscillation()
                block_size = int(math.ceil(block.size / dphi))
            elif block.units == "frames":
                block_size = int(math.ceil(block.size))
            else:
                raise RuntimeError(f"Unknown block_size unit {block.units!r}")
            if block_size > max_block_size:
                raise RuntimeError(
                    """
          The requested block size (%s) is larger than the maximum allowable block
          size (%s). Either decrease the requested block size or increase the
          amount of available memory.
        """
                    % (block_size, max_block_size)
                )
        block.size = block_size
        block.units = "frames"

    def compute_jobs(self):
        """
        Compute the jobs
        """
        imageset = self.experiments[0].imageset
        array_range = imageset.get_array_range()
        block = self.params.integration.block
        assert block.units == "frames"
        assert block.size > 0
        self.blocks = SimpleBlockList(array_range, block.size)
        assert len(self.blocks) > 0, "Invalid number of jobs"

    def summary(self):
        """
        Get a summary of the processing
        """
        # Compute the task table
        if self.experiments.all_stills():
            rows = [["#", "Group", "Frame From", "Frame To", "# Reflections"]]
            for i in range(len(self)):
                job = self.manager.job(i)
                group = job.index()
                f0, f1 = job.frames()
                n = self.manager.num_reflections(i)
                rows.append([str(i), str(group), str(f0), str(f1), str(n)])
        elif self.experiments.all_sequences():
            rows = [
                [
                    "#",
                    "Frame From",
                    "Frame To",
                    "Angle From",
                    "Angle To",
                    "# Reflections",
                ]
            ]
            for i in range(len(self)):
                f0, f1 = self.manager.job(i)
                scan = self.experiments[0].scan
                p0 = scan.get_angle_from_array_index(f0)
                p1 = scan.get_angle_from_array_index(f1)
                n = self.manager.num_reflections(i)
                rows.append([str(i), str(f0), str(f1), str(p0), str(p1), str(n)])
        else:
            raise RuntimeError("Experiments must be all sequences or all stills")

        # The job table
        task_table = tabulate(rows, headers="firstrow")

        # The format string
        if self.params.integration.block.size is None:
            block_size = "auto"
        else:
            block_size = str(self.params.integration.block.size)
        fmt = (
            "Processing reflections in the following blocks of images:\n"
            "\n"
            " block_size: %s %s\n"
            "\n"
            "%s\n"
        )
        return fmt % (block_size, self.params.integration.block.units, task_table)


class ReferenceCalculatorJob:
    """
    A class to represent an integration job
    """

    def __init__(self, index, job, experiments, reflections, params=None):
        """
        Initialise the task.

        :param index: The index of the processing job
        :param experiments: The list of experiments
        :param reflections: The list of reflections
        :param params: The processing parameters
        :param job: The frames to integrate
        :param flatten: Flatten the shoeboxes
        :param executor: The executor class
        """

        # Get the parameters
        if params is None:
            from dials.command_line.integrate import phil_scope

            params = phil_scope.extract()

        assert len(reflections) > 0, "Zero reflections given"
        assert (
            params.integration.block.max_memory_usage > 0.0
        ), "Max memory % must be > 0"
        assert (
            params.integration.block.max_memory_usage <= 1.0
        ), "Max memory % must be < 1"
        self.index = index
        self.job = job
        self.experiments = experiments
        self.reflections = reflections
        self.params = params

    def __call__(self):
        """
        Do the processing.

        :return: The processed data
        """
        from dials.algorithms.integration.processor import job

        # Set the global process ID
        job.index = self.index

        # Check all reflections have same imageset and get it
        imageset = self.experiments[0].imageset
        if not all(e.imageset == imageset for e in self.experiments):
            raise RuntimeError("Task can only handle 1 imageset")

        # Get the sub imageset
        frame0, frame1 = self.job
        try:
            frame10, frame11 = imageset.get_array_range()
        except Exception:
            frame10, frame11 = (0, len(imageset))
        try:
            assert frame0 < frame1
            assert frame10 < frame11
            assert frame0 >= frame10
            assert frame1 <= frame11
            index0 = frame0 - frame10
            index1 = index0 + (frame1 - frame0)
            assert index0 < index1
            assert index0 >= 0
            assert index1 <= len(imageset)
            imageset = imageset[index0:index1]
        except Exception:
            raise RuntimeError("Programmer Error: bad array range")

        # Check the memory requirements
        _assert_enough_memory(
            self.compute_required_memory(imageset),
            self.params.integration.block.max_memory_usage,
        )

        # Integrate
        self.compute_reference_profiles(imageset)

        # Write some debug files
        self.write_debug_files()

        # Return the result
        return dials.algorithms.integration.Result(
            index=self.index,
            reflections=self.reflections,
            data=self.reference,
            read_time=0,
            extract_time=0,
            process_time=0,
            total_time=0,
        )

    def compute_required_memory(self, imageset):
        return MultiThreadedIntegrator.compute_required_memory(
            imageset, self.params.integration.block.size
        )

    def compute_reference_profiles(self, imageset):
        """
        Integrate the reflections
        """
        from dials.algorithms.integration.integrator import frame_hist

        # Compute the partiality
        self.reflections.compute_partiality(self.experiments)

        # Get some info
        full_value = FULL_PARTIALITY - EPS
        fully_recorded = self.reflections["partiality"] > full_value
        npart = fully_recorded.count(False)
        nfull = fully_recorded.count(True)
        select_ice = self.reflections.get_flags(self.reflections.flags.in_powder_ring)
        select_int = ~self.reflections.get_flags(self.reflections.flags.dont_integrate)
        nice = select_ice.count(True)
        nint = select_int.count(True)
        ntot = len(self.reflections)
        frame0, frame1 = imageset.get_scan().get_array_range()

        # Write some output
        logger.info(" Beginning integration job %d", self.index)
        logger.info("")
        logger.info(" Frames: %d -> %d", frame0, frame1)
        logger.info("")
        logger.info(" Number of reflections")
        logger.info("  Partial:     %d", npart)
        logger.info("  Full:        %d", nfull)
        logger.info("  In ice ring: %d", nice)
        logger.info("  Integrate:   %d", nint)
        logger.info("  Total:       %d", ntot)
        logger.info("")

        # Print a histogram of reflections on frames
        if frame1 - frame0 > 1:
            logger.info(
                " The following histogram shows the number of reflections predicted"
            )
            logger.info(" to have all or part of their intensity on each frame.")
            logger.info("")
            logger.info(
                frame_hist(
                    self.reflections["bbox"].select(select_int), prefix=" ", symbol="*"
                )
            )
            logger.info("")

        # Construct the mask algorithm
        compute_mask = MaskCalculatorFactory.create(self.experiments, self.params)

        # Construct the background algorithm
        compute_background = BackgroundCalculatorFactory.create(
            self.experiments, self.params
        )

        # Construct the intensity algorithm
        compute_reference = ReferenceCalculatorFactory.create(
            self.experiments, self.params
        )

        # Call the multi threaded integrator
        reference_calculator = MultiThreadedReferenceProfiler(
            reflections=self.reflections,
            imageset=imageset,
            compute_mask=compute_mask,
            compute_background=compute_background,
            compute_reference=compute_reference,
            logger=Logger(logger),
            nthreads=self.params.integration.mp.nproc,
            buffer_size=self.params.integration.block.size,
            use_dynamic_mask=self.params.integration.use_dynamic_mask,
            debug=self.params.integration.debug.output,
        )

        # Assign the reflections
        self.reflections = reference_calculator.reflections()

        # Assign the reference profiles
        self.reference = compute_reference

        # Write some log output
        fmt = "Used %d / %d reflections to create reference profiles"
        dont_integrate = self.reflections.get_flags(
            self.reflections.flags.dont_integrate
        )
        used_in_modelling = self.reflections.get_flags(
            self.reflections.flags.used_in_modelling
        )
        n_tot = dont_integrate.count(False)
        n_mod = (used_in_modelling & ~dont_integrate).count(True)
        logger.info("")
        logger.info(fmt, n_mod, n_tot)

    def write_debug_files(self):
        """
        Write some debug output
        """

        # Optionally save the shoeboxes
        debug = self.params.integration.debug
        if debug.output and debug.separate_files:
            output = self.reflections
            if debug.select is not None:
                output = output.select(debug.select(output))
            if debug.split_experiments:
                output = output.split_by_experiment_id()
                for table in output:
                    i = table["id"][0]
                    table.as_file("shoeboxes_%d_%d.refl" % (self.index, i))
            else:
                output.as_file("shoeboxes_%d.refl" % self.index)

        # Delete the shoeboxes
        if debug.separate_files or not debug.output:
            del self.reflections["shoebox"]


class ReferenceCalculatorManager:
    """
    A class to manage processing book-keeping
    """

    def __init__(self, experiments, reflections, params):
        """
        Initialise the manager.

        :param experiments: The list of experiments
        :param reflections: The list of reflections
        :param params: The phil parameters
        """

        # Save some data
        self.experiments = experiments
        self.reflections = reflections
        self.reference = None

        # Save some parameters
        self.params = params

        # Set the finalized flag to False
        self.finalized = False

        # Initialise the timing information
        # self.time = TimingInfo()

        self.initialize()

    def initialize(self):
        """
        Initialise the processing
        """
        # Ensure the reflections contain bounding boxes
        assert "bbox" in self.reflections, "Reflections have no bbox"

        # Select only those reflections used in refinement
        selection = self.reflections.get_flags(self.reflections.flags.reference_spot)
        if selection.count(True) == 0:
            raise RuntimeError("No reference reflections given")
        self.reflections = self.reflections.select(selection)

        # Compute the block size and jobs
        self.compute_blocks()
        self.compute_jobs()
        self.reflections = split_partials_over_boundaries(
            self.reflections, self.params.integration.block.size
        )

        # Create the reflection manager
        self.manager = SimpleReflectionManager(
            self.blocks, self.reflections, self.params.integration.mp.njobs
        )

    def task(self, index):
        """
        Get a task.
        """
        frames = self.manager.job(index)
        experiments = self.experiments
        reflections = self.manager.split(index)
        if len(reflections) == 0:
            logger.warning("No reflections in job %d ***", index)
            task = NullTask(index=index, reflections=reflections)
        else:
            task = ReferenceCalculatorJob(
                index=index,
                job=frames,
                experiments=experiments,
                reflections=reflections,
                params=self.params,
            )
        return task

    def tasks(self):
        """
        Iterate through the tasks.
        """
        for i in range(len(self)):
            yield self.task(i)

    def accumulate(self, result):
        """Accumulate the results."""
        self.manager.accumulate(result.index, result.reflections)

        if self.reference is None:
            self.reference = result.data
        else:
            self.reference.accumulate(result.data)

    def finalize(self):
        """
        Finalize the processing and finish.
        """
        # Check manager is finished
        assert self.manager.finished(), "Manager is not finished"

        # Set the reference profiles
        self.reference = self.reference.reference_profiles()

        self.finalized = True

    def result(self):
        """
        Return the result.

        :return: The result
        """
        assert self.finalized, "Manager is not finalized"
        return self.manager.data()

    def finished(self):
        """
        Return if all tasks have finished.

        :return: True/False all tasks have finished
        """
        return self.finalized and self.manager.finished()

    def __len__(self):
        """
        Return the number of tasks.

        :return: the number of tasks
        """
        return len(self.manager)

    def compute_max_block_size(self):
        """
        Compute the required memory
        """
        total_memory = psutil.virtual_memory().available
        max_memory_usage = self.params.integration.block.max_memory_usage
        assert max_memory_usage > 0.0, "maximum memory usage must be > 0"
        assert max_memory_usage <= 1.0, "maximum memory usage must be <= 1"
        limit_memory = int(math.floor(total_memory * max_memory_usage))
        return MultiThreadedReferenceProfiler.compute_max_block_size(
            self.experiments[0].imageset, max_memory_usage=limit_memory
        )

    def compute_blocks(self):
        """
        Compute the processing block size.
        """
        block = self.params.integration.block
        max_block_size = self.compute_max_block_size()
        if block.size in [Auto, "auto", "Auto"]:
            assert block.threshold > 0, "Threshold must be > 0"
            assert block.threshold <= 1.0, "Threshold must be < 1"
            nframes = sorted([b[5] - b[4] for b in self.reflections["bbox"]])
            cutoff = int(block.threshold * len(nframes))
            block_size = nframes[cutoff]
            if block_size > max_block_size:
                logger.warning(
                    "Computed block size (%s) > maximum block size (%s).",
                    block_size,
                    max_block_size,
                )
                logger.warning(
                    "Setting block size to maximum; some reflections may be partial"
                )
                block_size = max_block_size
        else:
            scan = self.experiments[0].scan
            if block.units == "radians":
                phi0, dphi = scan.get_oscillation(deg=False)
                block_size = int(math.ceil(block.size / dphi))
            elif block.units == "degrees":
                phi0, dphi = scan.get_oscillation()
                block_size = int(math.ceil(block.size / dphi))
            elif block.units == "frames":
                block_size = int(math.ceil(block.size))
            else:
                raise RuntimeError(f"Unknown block_size unit {block.units!r}")
            if block_size > max_block_size:
                raise RuntimeError(
                    """
          The requested block size (%s) is larger than the maximum allowable block
          size (%s). Either decrease the requested block size or increase the
          amount of available memory.
        """
                    % (block_size, max_block_size)
                )
        block.size = block_size
        block.units = "frames"

    def compute_jobs(self):
        imageset = self.experiments[0].imageset
        array_range = imageset.get_array_range()
        block = self.params.integration.block
        assert block.units == "frames"
        assert block.size > 0
        self.blocks = SimpleBlockList(array_range, block.size)
        assert len(self.blocks) > 0, "Invalid number of jobs"

    def summary(self):
        """
        Get a summary of the processing
        """
        # Compute the task table
        if self.experiments.all_stills():
            rows = [["#", "Group", "Frame From", "Frame To", "# Reflections"]]
            for i in range(len(self)):
                job = self.manager.job(i)
                group = job.index()
                f0, f1 = job.frames()
                n = self.manager.num_reflections(i)
                rows.append([str(i), str(group), str(f0), str(f1), str(n)])
        elif self.experiments.all_sequences():
            rows = [
                [
                    "#",
                    "Frame From",
                    "Frame To",
                    "Angle From",
                    "Angle To",
                    "# Reflections",
                ]
            ]
            for i in range(len(self)):
                f0, f1 = self.manager.job(i)
                scan = self.experiments[0].scan
                p0 = scan.get_angle_from_array_index(f0)
                p1 = scan.get_angle_from_array_index(f1)
                n = self.manager.num_reflections(i)
                rows.append([str(i), str(f0), str(f1), str(p0), str(p1), str(n)])
        else:
            raise RuntimeError("Experiments must be all sequences or all stills")

        # The job table
        task_table = tabulate(rows, headers="firstrow")

        # The format string
        if self.params.integration.block.size is None:
            block_size = "auto"
        else:
            block_size = str(self.params.integration.block.size)
        fmt = (
            "Processing reflections in the following blocks of images:\n"
            "\n"
            " block_size: %s %s\n"
            "\n"
            "%s\n"
        )
        return fmt % (block_size, self.params.integration.block.units, task_table)


def compute_required_memory(imageset, block_size):
    """
    Compute the required memory

    """
    return MultiThreadedIntegrator.compute_required_memory(imageset, block_size)


class ReferenceCalculatorProcessor:
    def __init__(self, experiments, reflections, params=None):
        from dials.util import pprint

        # Create the reference manager
        reference_manager = ReferenceCalculatorManager(experiments, reflections, params)

        # Print some output
        logger.info(reference_manager.summary())

        # Execute each task
        if params.integration.mp.njobs > 1:

            if params.integration.mp.method == "multiprocessing":
                _assert_enough_memory(
                    params.integration.mp.njobs
                    * compute_required_memory(
                        experiments[0].imageset, params.integration.block.size
                    ),
                    params.integration.block.max_memory_usage,
                )

            def process_output(result):
                for message in result[1]:
                    logger.log(message.levelno, message.msg)
                reference_manager.accumulate(result[0])

            multi_node_parallel_map(
                func=execute_parallel_task,
                iterable=reference_manager.tasks(),
                nproc=params.integration.mp.nproc,
                njobs=params.integration.mp.njobs,
                callback=process_output,
                cluster_method=params.integration.mp.method,
                preserve_order=True,
            )
        else:
            for task in reference_manager.tasks():
                result = task()
                reference_manager.accumulate(result)

        # Finalize the processing
        reference_manager.finalize()

        # Set the reflections and profiles
        self._reflections = reference_manager.result()
        self._profiles = reference_manager.reference

        # Write the profiles to file
        if params.integration.debug.reference.output:
            with open(params.integration.debug.reference.filename, "wb") as outfile:
                pickle.dump(self._profiles, outfile)

        # Print the profiles to the debug log
        for i in range(len(self._profiles)):
            logger.debug("")
            logger.debug("Reference Profiles for experiment %d", i)
            logger.debug("")
            reference = self._profiles[i].reference()
            for j in range(len(reference)):
                data = reference.data(j)
                logger.debug("Profile %d", j)
                if len(data) > 0:
                    logger.debug(pprint.profile3d(data))
                else:
                    logger.debug("** NO PROFILE **")

    def reflections(self):
        return self._reflections

    def profiles(self):
        return self._profiles


class IntegratorProcessor:
    def __init__(self, experiments, reflections, reference=None, params=None):

        # Create the reference manager
        integration_manager = IntegrationManager(
            experiments, reflections, reference, params
        )

        # Print some output
        logger.info(integration_manager.summary())

        # Execute each task
        if params.integration.mp.njobs > 1:

            if params.integration.mp.method == "multiprocessing":
                _assert_enough_memory(
                    params.integration.mp.njobs
                    * compute_required_memory(
                        experiments[0].imageset, params.integration.block.size
                    ),
                    params.integration.block.max_memory_usage,
                )

            def process_output(result):
                for message in result[1]:
                    logger.log(message.levelno, message.msg)
                integration_manager.accumulate(result[0])

            multi_node_parallel_map(
                func=execute_parallel_task,
                iterable=integration_manager.tasks(),
                nproc=params.integration.mp.nproc,
                njobs=params.integration.mp.njobs,
                callback=process_output,
                cluster_method=params.integration.mp.method,
                preserve_order=True,
            )
        else:
            for task in integration_manager.tasks():
                result = task()
                integration_manager.accumulate(result)

        # Finalize the processing
        integration_manager.finalize()

        # Set the reflections and profiles
        self._reflections = integration_manager.result()

    def reflections(self):
        return self._reflections


def split_partials_over_boundaries(reflections, block_size):
    """
    Split the reflections into partials or over job boundaries
    """

    # Get the block size and num frames
    _, _, _, _, z0, z1 = reflections["bbox"].parts()
    n_frames_of_bboxes = z1 - z0
    num_full = len(reflections)

    # See if any reflections need to be split
    refl_to_split_sel = n_frames_of_bboxes > block_size
    if refl_to_split_sel.count(True) > 0:

        # Get the subset of reflections to be split
        subset = reflections.select(refl_to_split_sel)
        newset = flex.reflection_table()
        for i in range(len(subset)):
            # get a copy of the reflection using selection
            item = subset.select(flex.size_t([i]))
            bbox = item["bbox"][0]
            size = bbox[5] - bbox[4]
            assert size > block_size
            nsplits = int(math.ceil(float(size) / float(block_size)))
            partsize = int(math.ceil(float(size) / float(nsplits)))
            bbox0 = bbox[0:4] + (bbox[4], bbox[4] + partsize)
            subset["bbox"][i] = bbox0  # set the updated bbox in the subset
            for _ in range(1, nsplits):
                # make z0 start at end of previous bbox0
                bbox0 = bbox0[0:4] + (bbox0[5], min(bbox0[5] + partsize, bbox[5]))
                assert bbox0[4] < bbox0[5]
                assert bbox0[5] <= bbox[5]
                # set the split bbox as the updated bbox in the reflection copy
                item["bbox"][0] = bbox0
                # and add to newset as separate reflection
                newset.extend(item)

        # Update the subset of trimmed reflections and add the new set of partials
        reflections.set_selected(refl_to_split_sel, subset)
        reflections.extend(newset)

    # Print some info
    num_after_splitting = len(reflections)
    assert num_after_splitting >= num_full, "Invalid number of partials"
    if num_after_splitting > num_full:
        num_split = num_after_splitting - num_full
        logger.info(" Split %d reflections\n", num_split)

    return reflections



from __future__ import annotations

import os

import procrunner


def test_export_single_bitmap(dials_data, tmp_path):
    result = procrunner.run(
        [
            "dials.export_bitmaps",
            dials_data("centroid_test_data", pathlib=True) / "centroid_0001.cbf",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    assert tmp_path.joinpath("image0001.png").is_file()


def test_export_multiple_bitmaps(dials_data, tmp_path):
    result = procrunner.run(
        [
            "dials.export_bitmaps",
            dials_data("centroid_test_data", pathlib=True) / "experiments.json",
            "prefix=variance_",
            "binning=2",
            "display=variance",
            "colour_scheme=inverse_greyscale",
            "brightness=25",
            "kernel_size=5,5",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr

    for i in range(1, 8):
        assert tmp_path.joinpath(f"variance_000{i}.png").is_file()


def test_export_bitmap_with_prefix_and_no_padding(dials_data, tmp_path):
    result = procrunner.run(
        [
            "dials.export_bitmaps",
            dials_data("centroid_test_data", pathlib=True) / "centroid_0001.cbf",
            "prefix=img_",
            "padding=0",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    assert tmp_path.joinpath("img_1.png").is_file()


def test_export_bitmap_with_prefix_and_extra_padding(dials_data, tmp_path):
    result = procrunner.run(
        [
            "dials.export_bitmaps",
            dials_data("centroid_test_data", pathlib=True) / "centroid_0001.cbf",
            "prefix=img_",
            "padding=5",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    assert tmp_path.joinpath("img_00001.png").is_file()


def test_export_bitmap_with_specified_output_filename(dials_data, tmp_path):
    result = procrunner.run(
        [
            "dials.export_bitmaps",
            dials_data("centroid_test_data", pathlib=True) / "centroid_0001.cbf",
            "output.file=kittens.png",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    assert tmp_path.joinpath("kittens.png").is_file()


def test_export_multiple_bitmaps_with_specified_output_filename_fails(
    dials_data, tmp_path
):
    # setting output filename not allowed with >1 image
    result = procrunner.run(
        [
            "dials.export_bitmaps",
            dials_data("centroid_test_data", pathlib=True) / "experiments.json",
            "output.file=kittens.png",
        ],
        working_directory=tmp_path,
    )
    assert result.returncode


def test_export_still_image(dials_regression, tmp_path):
    image = os.path.join(
        dials_regression, "image_examples", "DLS_I24_stills", "still_0001.cbf"
    )

    result = procrunner.run(["dials.export_bitmaps", image], working_directory=tmp_path)
    assert not result.returncode and not result.stderr

    assert tmp_path.joinpath("image0001.png").is_file()


def test_export_multi_panel(dials_regression, tmp_path):
    image = os.path.join(
        dials_regression, "image_examples", "DLS_I23", "germ_13KeV_0001.cbf"
    )

    for binning in (1, 4):
        result = procrunner.run(
            [
                "dials.export_bitmaps",
                image,
                "binning=%i" % binning,
                "prefix=binning_%i_" % binning,
            ],
            working_directory=tmp_path,
        )
        assert not result.returncode and not result.stderr
        assert tmp_path.joinpath(f"binning_{binning}_0001.png").is_file()


def test_export_restricted_multiimage(dials_data, tmp_path):
    "Test exporting a subset of an imageset"
    result = procrunner.run(
        [
            "dials.export_bitmaps",
            dials_data("centroid_test_data", pathlib=True) / "experiments.json",
            "imageset_index=2",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    assert [f.name for f in tmp_path.glob("*.png")] == [
        "image0002.png"
    ], "Only one image expected"



from __future__ import annotations

import py.path
import pytest

from cctbx import sgtbx, uctbx
from dxtbx.imageset import ImageSet
from dxtbx.serialize import load

from dials.algorithms.indexing import lattice_search, stills_indexer
from dials.array_family import flex
from dials.command_line.index import phil_scope
from dials.command_line.slice_sequence import slice_experiments, slice_reflections


@pytest.fixture
def i04_weak_data(dials_regression):
    data_dir = py.path.local(dials_regression).join(
        "indexing_test_data", "i04_weak_data"
    )
    reflections_path = data_dir.join("full.pickle").strpath
    experiments_path = data_dir.join("experiments_import.json").strpath

    reflections = flex.reflection_table.from_file(reflections_path)
    experiments = load.experiment_list(experiments_path, check_format=False)

    return {
        "reflections": slice_reflections(reflections, [(1, 20)]),
        "experiments": slice_experiments(experiments, [(1, 20)]),
    }


@pytest.mark.parametrize(
    "indexing_method,space_group,unit_cell",
    (
        ("fft1d", None, None),
        ("fft3d", None, None),
        ("real_space_grid_search", "P422", (57.8, 57.8, 150.0, 90, 90, 90)),
    ),
)
def test_BasisVectorSearch_i04_weak_data(
    i04_weak_data, indexing_method, space_group, unit_cell
):
    reflections = i04_weak_data["reflections"]
    experiments = i04_weak_data["experiments"]
    params = phil_scope.fetch().extract()
    params.indexing.refinement_protocol.n_macro_cycles = 2
    params.indexing.basis_vector_combinations.max_refine = 5
    params.indexing.method = indexing_method
    if unit_cell is not None:
        params.indexing.known_symmetry.unit_cell = uctbx.unit_cell(unit_cell)
    if space_group is not None:
        params.indexing.known_symmetry.space_group = sgtbx.space_group_info(space_group)
    idxr = lattice_search.BasisVectorSearch(reflections, experiments, params)
    idxr.index()

    indexed_experiments = idxr.refined_experiments
    assert len(indexed_experiments) == 1
    assert indexed_experiments[0].crystal.get_unit_cell().parameters() == pytest.approx(
        (57.752, 57.776, 150.013, 90.0101, 89.976, 90.008), rel=1e-3
    )


@pytest.mark.xfel
@pytest.mark.parametrize(
    "indexing_method,space_group,unit_cell",
    (
        ("fft3d", "P422", (57.8, 57.8, 150.0, 90, 90, 90)),
        ("fft1d", "P422", (57.8, 57.8, 150.0, 90, 90, 90)),
        ("real_space_grid_search", "P422", (57.8, 57.8, 150.0, 90, 90, 90)),
        ("low_res_spot_match", "P422", (57.8, 57.8, 150.0, 90, 90, 90)),
    ),
)
def test_stills_indexer_methods_i04_weak_data(
    i04_weak_data, indexing_method, space_group, unit_cell
):
    reflections = slice_reflections(i04_weak_data["reflections"], [(1, 2)])
    experiments = slice_experiments(i04_weak_data["experiments"], [(1, 2)])
    for experiment in experiments:
        experiment.imageset = ImageSet(
            experiment.imageset.data(), experiment.imageset.indices()
        )
        experiment.imageset.set_scan(None)
        experiment.imageset.set_goniometer(None)
        experiment.scan = None
        experiment.goniometer = None
    params = phil_scope.fetch().extract()
    params.indexing.method = indexing_method
    params.indexing.basis_vector_combinations.max_refine = 5
    if unit_cell is not None:
        params.indexing.known_symmetry.unit_cell = uctbx.unit_cell(unit_cell)
    if space_group is not None:
        params.indexing.known_symmetry.space_group = sgtbx.space_group_info(space_group)
    try:
        idxr = stills_indexer.StillsIndexerBasisVectorSearch(
            reflections, experiments, params
        )
    except RuntimeError:
        idxr = stills_indexer.StillsIndexerLatticeSearch(
            reflections, experiments, params
        )
    idxr.index()

    indexed_experiments = idxr.refined_experiments
    assert len(indexed_experiments) == 1
    assert indexed_experiments[0].crystal.get_unit_cell().parameters() == pytest.approx(
        (57.752, 57.776, 150.013, 90.0101, 89.976, 90.008), rel=1e-2
    )



from __future__ import annotations

import random
import textwrap
from math import cos, pi, sin, sqrt

from dxtbx.model import BeamFactory, Detector, DetectorFactory, Panel
from libtbx.test_utils import approx_equal
from scitbx import matrix

from dials.algorithms.refinement.parameterisation.detector_parameters import (
    DetectorParameterisationMultiPanel,
    DetectorParameterisationSinglePanel,
)
from dials.algorithms.refinement.refinement_helpers import (
    get_fd_gradients,
    random_param_shift,
)


def random_panel(lim=(0, 50)):
    """For testing, return a square panel with a randomised position
    and orientation"""

    # start with a randomised origin vector
    o = matrix.col(
        (
            random.uniform(-200, 200),
            random.uniform(-200, 200),
            random.uniform(-200, 200),
        )
    )

    # two orthogonal unit vectors randomly oriented in the normal plane
    # of the origin vector
    u1 = o.ortho().normalize()
    u2 = o.cross(u1).normalize()
    # theta = random.uniform(0, 2. * pi)
    u1 = u1.rotate_around_origin(o, pi / 12)
    u2 = u2.rotate_around_origin(o, pi / 12)

    # offset the plane normal from the origin vector by random rotations
    # of up to 45 degrees for each direction
    u1 = u1.rotate_around_origin(u2, random.uniform(-pi / 2.0, pi / 2.0))
    u2 = u2.rotate_around_origin(u1, random.uniform(-pi / 2.0, pi / 2.0))

    return Panel(
        "PAD",
        "Panel",
        u1,
        u2,
        o,
        (lim[1] / 200, lim[1] / 200),
        (200, 200),
        (0, 2e20),
        0.0,
        "",
    )


# local function required to make a 3x3 multi-panel detector
def make_multi_panel(single_panel_detector):
    """Create a 3x3 multi-panel detector filling the same space as
    a supplied single panel detector"""

    from .setup_geometry import random_vector_close_to
    from .test_multi_panel_detector_parameterisation import make_panel_in_array

    multi_panel_detector = Detector()
    for x in range(3):
        for y in range(3):
            new_panel = make_panel_in_array((x, y), single_panel_detector[0])
            multi_panel_detector.add_panel(new_panel)

    # apply small random shifts & rotations to each panel
    for p in multi_panel_detector:

        # perturb origin vector
        o_multiplier = random.gauss(1.0, 0.01)
        new_origin = random_vector_close_to(p.get_origin(), sd=0.1)
        new_origin *= o_multiplier

        # perturb fast direction vector
        new_dir1 = random_vector_close_to(p.get_fast_axis(), sd=0.5)

        # create vector in the plane of dir1-dir2
        dir1_dir2 = random_vector_close_to(p.get_slow_axis(), sd=0.5)

        # find normal to panel plane and thus new slow direction vector
        dn = new_dir1.cross(dir1_dir2)
        new_dir2 = dn.cross(new_dir1)

        # set panel frame
        p.set_frame(new_dir1, new_dir2, new_origin)

    return multi_panel_detector


def test():
    # set the random seed to make the test reproducible
    random.seed(1337)

    # set up a simple detector frame with directions aligned with
    # principal axes and sensor origin located on the z-axis at -110
    d1 = matrix.col((1, 0, 0))
    d2 = matrix.col((0, -1, 0))
    # lim = (0,50)
    npx_fast = 1475
    npx_slow = 1679
    pix_size_f = pix_size_s = 0.172
    detector = DetectorFactory.make_detector(
        "PAD",
        d1,
        d2,
        matrix.col((0, 0, -110)),
        (pix_size_f, pix_size_s),
        (npx_fast, npx_slow),
        (0, 2e20),
    )

    dp = DetectorParameterisationSinglePanel(detector)
    beam = BeamFactory().make_beam(
        sample_to_source=-1 * (matrix.col((0, 0, -110)) + 10 * d1 + 10 * d2),
        wavelength=1.0,
    )

    # Test change of parameters
    # =========================

    # 1. shift detector plane so that the z-axis intercepts its centre
    # at a distance of 100 along the initial normal direction. As the
    # initial normal is along -z, we expect the frame to intercept the
    # z-axis at -100.

    p_vals = dp.get_param_vals()
    p_vals[0:3] = [100.0, 0.0, 0.0]
    dp.set_param_vals(p_vals)
    detector = dp._model
    assert len(detector) == 1
    panel = detector[0]
    v1 = matrix.col(panel.get_origin())
    v2 = matrix.col((0.0, 0.0, 1.0))
    assert approx_equal(v1.dot(v2), -100.0)

    # 2. rotate frame around its initial normal by +90 degrees. Only d1
    # and d2 should change. As we rotate clockwise around the initial
    # normal (-z direction) then d1 should rotate onto the original
    # direction d2, and d2 should rotate to negative of the original
    # direction d1

    p_vals[3] = 1000.0 * pi / 2  # set tau1 value
    dp.set_param_vals(p_vals)

    detector = dp._model
    assert len(detector) == 1
    panel = detector[0]
    assert approx_equal(
        matrix.col(panel.get_fast_axis()).dot(dp._initial_state["d1"]), 0.0
    )
    assert approx_equal(
        matrix.col(panel.get_slow_axis()).dot(dp._initial_state["d2"]), 0.0
    )
    assert approx_equal(
        matrix.col(panel.get_normal()).dot(dp._initial_state["dn"]), 1.0
    )

    # 3. no rotation around initial normal, +10 degrees around initial
    # d1 direction and +10 degrees around initial d2. Check d1 and d2
    # match paper calculation

    p_vals[3] = 0.0  # tau1
    p_vals[4] = 1000.0 * pi / 18  # tau2
    p_vals[5] = 1000.0 * pi / 18  # tau3
    dp.set_param_vals(p_vals)

    # paper calculation values
    v1 = matrix.col((cos(pi / 18), 0, sin(pi / 18)))
    v2 = matrix.col(
        (
            sin(pi / 18) ** 2,
            -cos(pi / 18),
            sqrt((2 * sin(pi / 36) * sin(pi / 18)) ** 2 - sin(pi / 18) ** 4)
            - sin(pi / 18),
        )
    )

    detector = dp._model
    assert len(detector) == 1
    panel = detector[0]
    assert approx_equal(matrix.col(panel.get_fast_axis()).dot(v1), 1.0)
    assert approx_equal(matrix.col(panel.get_slow_axis()).dot(v2), 1.0)

    # 4. Test fixing and unfixing of parameters
    p_vals = [100.0, 0.0, 0.0, 1000.0 * pi / 18, 1000.0 * pi / 18, 1000.0 * pi / 18]
    dp.set_param_vals(p_vals)
    f = dp.get_fixed()
    f[0:3] = [True] * 3
    dp.set_fixed(f)
    p_vals2 = [0.0, 0.0, 0.0]
    dp.set_param_vals(p_vals2)
    assert dp.get_param_vals(only_free=False) == [100.0, 0.0, 0.0, 0.0, 0.0, 0.0]

    an_ds_dp = dp.get_ds_dp()
    assert len(an_ds_dp) == 3

    f[0:3] = [False] * 3
    dp.set_fixed(f)
    p_vals = dp.get_param_vals()
    p_vals2 = [a + b for a, b in zip(p_vals, [-10.0, 1.0, 1.0, 0.0, 0.0, 0.0])]
    dp.set_param_vals(p_vals2)
    assert dp.get_param_vals() == [90.0, 1.0, 1.0, 0.0, 0.0, 0.0]

    # 5. Tests of the calculation of derivatives

    # Now using parameterisation in mrad

    # random initial orientations with a random parameter shift at each
    attempts = 100
    for i in range(attempts):

        # create random initial position
        det = Detector(random_panel())
        dp = DetectorParameterisationSinglePanel(det)

        # apply a random parameter shift
        p_vals = dp.get_param_vals()
        p_vals = random_param_shift(
            p_vals, [10, 10, 10, 1000.0 * pi / 18, 1000.0 * pi / 18, 1000.0 * pi / 18]
        )
        dp.set_param_vals(p_vals)

        # compare analytical and finite difference derivatives.
        an_ds_dp = dp.get_ds_dp(multi_state_elt=0)
        fd_ds_dp = get_fd_gradients(dp, [1.0e-6] * 3 + [1.0e-4 * pi / 180] * 3)

        for j in range(6):
            assert approx_equal(
                (fd_ds_dp[j] - an_ds_dp[j]),
                matrix.sqr((0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0)),
                eps=1.0e-6,
            ), textwrap.dedent(
                """\
        Failure comparing analytical with finite difference derivatives.
        Failure in try {i}
        failure for parameter number {j}
        of the orientation parameterisation
        with fd_ds_dp =
        {fd}
        and an_ds_dp =
        {an}
        so that difference fd_ds_dp - an_ds_dp =
        {diff}
        """
            ).format(
                i=i, j=j, fd=fd_ds_dp[j], an=an_ds_dp[j], diff=fd_ds_dp[j] - an_ds_dp[j]
            )

    # 5. Test a multi-panel detector with non-coplanar panels.

    # place a beam at the centre of the single panel detector (need a
    # beam to initialise the multi-panel detector parameterisation)
    lim = det[0].get_image_size_mm()
    shift1 = lim[0] / 2.0
    shift2 = lim[1] / 2.0
    beam_centre = (
        matrix.col(det[0].get_origin())
        + shift1 * matrix.col(det[0].get_fast_axis())
        + shift2 * matrix.col(det[0].get_slow_axis())
    )
    beam = BeamFactory().make_beam(sample_to_source=-1.0 * beam_centre, wavelength=1.0)

    multi_panel_detector = make_multi_panel(det)

    # parameterise this detector
    dp = DetectorParameterisationMultiPanel(multi_panel_detector, beam)

    # ensure the beam still intersects the central panel
    intersection = multi_panel_detector.get_ray_intersection(beam.get_s0())
    assert intersection[0] == 4

    # record the offsets and dir1s, dir2s
    offsets_before_shift = dp._offsets
    dir1s_before_shift = dp._dir1s
    dir2s_before_shift = dp._dir2s

    # apply a random parameter shift (~10 mm distances, ~50 mrad angles)
    p_vals = dp.get_param_vals()
    p_vals = random_param_shift(p_vals, [10, 10, 10, 50, 50, 50])

    # reparameterise the detector
    dp = DetectorParameterisationMultiPanel(multi_panel_detector, beam)

    # record the offsets and dir1s, dir2s
    offsets_after_shift = dp._offsets
    dir1s_after_shift = dp._dir1s
    dir2s_after_shift = dp._dir2s

    # ensure the offsets, dir1s and dir2s are the same. This means that
    # each panel in the detector moved with the others as a rigid body
    for a, b in zip(offsets_before_shift, offsets_after_shift):
        assert approx_equal(a, b, eps=1.0e-10)

    for a, b in zip(dir1s_before_shift, dir1s_after_shift):
        assert approx_equal(a, b, eps=1.0e-10)

    for a, b in zip(dir2s_before_shift, dir2s_after_shift):
        assert approx_equal(a, b, eps=1.0e-10)

    attempts = 5
    for i in range(attempts):

        multi_panel_detector = make_multi_panel(det)

        # parameterise this detector
        dp = DetectorParameterisationMultiPanel(multi_panel_detector, beam)
        p_vals = dp.get_param_vals()

        # apply a random parameter shift
        p_vals = random_param_shift(
            p_vals, [10, 10, 10, 1000.0 * pi / 18, 1000.0 * pi / 18, 1000.0 * pi / 18]
        )
        dp.set_param_vals(p_vals)

        # compare analytical and finite difference derivatives
        # get_fd_gradients will implicitly only get gradients for the
        # 1st panel in the detector, so explicitly get the same for the
        # analytical gradients

        for j in range(9):

            an_ds_dp = dp.get_ds_dp(multi_state_elt=j)
            fd_ds_dp = get_fd_gradients(dp, [1.0e-7] * dp.num_free(), multi_state_elt=j)

            for k in range(6):
                assert approx_equal(
                    (fd_ds_dp[k] - matrix.sqr(an_ds_dp[k])),
                    matrix.sqr((0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0)),
                    eps=1.0e-5,
                    out=None,
                ), textwrap.dedent(
                    """\
        Failure comparing analytical with finite difference derivatives.
        Failure in try {i}
        for panel number {j]
        failure for parameter number {k}
        of the orientation parameterisation
        with fd_ds_dp =
        {fd}
        and an_ds_dp =
        {an}
        so that difference fd_ds_dp - an_ds_dp =
        {diff}
        """
                ).format(
                    i=i,
                    j=j,
                    k=k,
                    fd=fd_ds_dp[k],
                    an=an_ds_dp[k],
                    diff=fd_ds_dp[k] - matrix.sqr(an_ds_dp[k]),
                )



from __future__ import annotations

import math
from random import randint, seed

import numpy as np

from dxtbx.model.experiment_list import ExperimentList
from scitbx import matrix

from dials.algorithms.profile_model.gaussian_rs import (
    CoordinateSystem,
    MaskCalculator3D,
)
from dials.algorithms.shoebox import MaskCode
from dials.array_family import flex


def test(dials_data):
    experiment = ExperimentList.from_file(
        dials_data("centroid_test_data", pathlib=True) / "experiments.json"
    )

    beam = experiment[0].beam
    detector = experiment[0].detector
    goniometer = experiment[0].goniometer
    scan = experiment[0].scan
    delta_b = experiment[0].profile.delta_b()
    delta_m = experiment[0].profile.delta_m()

    assert len(detector) == 1

    # Get the function object to mask the foreground
    mask_foreground = MaskCalculator3D(
        beam, detector, goniometer, scan, delta_b, delta_m
    )

    s0 = beam.get_s0()
    m2 = goniometer.get_rotation_axis()
    s0_length = matrix.col(beam.get_s0()).length()
    width, height = detector[0].get_image_size()
    zrange = scan.get_array_range()
    phi0, dphi = scan.get_oscillation(deg=False)

    # Generate some reflections
    reflections = generate_reflections(detector, beam, scan, experiment, 10)

    # Mask the foreground in each
    mask_foreground(
        reflections["shoebox"],
        reflections["s1"],
        reflections["xyzcal.px"].parts()[2],
        reflections["panel"],
    )

    # Loop through all the reflections and check the mask values
    shoebox = reflections["shoebox"]
    beam_vector = reflections["s1"]
    rotation_angle = reflections["xyzcal.mm"].parts()[2]
    for l in range(len(reflections)):
        mask = shoebox[l].mask
        x0, x1, y0, y1, z0, z1 = shoebox[l].bbox
        s1 = beam_vector[l]
        phi = rotation_angle[l]
        cs = CoordinateSystem(m2, s0, s1, phi)

        def rs_coord(i, j, k):
            s1d = detector[0].get_pixel_lab_coord((i, j))
            s1d = matrix.col(s1d).normalize() * s0_length
            e1, e2 = cs.from_beam_vector(s1d)
            e3 = cs.from_rotation_angle_fast(phi0 + (k - zrange[0]) * dphi)
            return e1, e2, e3

        new_mask = flex.int(mask.accessor(), 0)
        for k in range(z1 - z0):
            for j in range(y1 - y0):
                for i in range(x1 - x0):
                    # value1 = mask[k, j, i]
                    e11, e12, e13 = rs_coord(x0 + i, y0 + j, z0 + k)
                    e21, e22, e23 = rs_coord(x0 + i + 1, y0 + j, z0 + k)
                    e31, e32, e33 = rs_coord(x0 + i, y0 + j + 1, z0 + k)
                    e41, e42, e43 = rs_coord(x0 + i, y0 + j, z0 + k + 1)
                    e51, e52, e53 = rs_coord(x0 + i + 1, y0 + j + 1, z0 + k)
                    e61, e62, e63 = rs_coord(x0 + i + 1, y0 + j, z0 + k + 1)
                    e71, e72, e73 = rs_coord(x0 + i, y0 + j + 1, z0 + k + 1)
                    e81, e82, e83 = rs_coord(x0 + i + 1, y0 + j + 1, z0 + k + 1)
                    de1 = (e11 / delta_b) ** 2 + (
                        e12 / delta_b
                    ) ** 2  # +(e13/delta_m)**2
                    de2 = (e21 / delta_b) ** 2 + (
                        e22 / delta_b
                    ) ** 2  # +(e23/delta_m)**2
                    de3 = (e31 / delta_b) ** 2 + (
                        e32 / delta_b
                    ) ** 2  # +(e33/delta_m)**2
                    de4 = (e41 / delta_b) ** 2 + (
                        e42 / delta_b
                    ) ** 2  # +(e43/delta_m)**2
                    de5 = (e51 / delta_b) ** 2 + (
                        e52 / delta_b
                    ) ** 2  # +(e53/delta_m)**2
                    de6 = (e61 / delta_b) ** 2 + (
                        e62 / delta_b
                    ) ** 2  # +(e63/delta_m)**2
                    de7 = (e71 / delta_b) ** 2 + (
                        e72 / delta_b
                    ) ** 2  # +(e73/delta_m)**2
                    de8 = (e81 / delta_b) ** 2 + (
                        e82 / delta_b
                    ) ** 2  # +(e83/delta_m)**2
                    de = math.sqrt(min([de1, de2, de3, de4, de5, de6, de7, de8]))
                    if (
                        x0 + i < 0
                        or y0 + j < 0
                        or x0 + i >= width
                        or y0 + j >= height
                        or z0 + k < zrange[0]
                        or z0 + k >= zrange[1]
                    ):
                        value2 = MaskCode.Valid
                    else:
                        if de <= 1.0:
                            value2 = MaskCode.Valid | MaskCode.Foreground
                        else:
                            value2 = MaskCode.Valid | MaskCode.Background
                    new_mask[k, j, i] = value2

        if not all(m1 == m2 for m1, m2 in zip(mask, new_mask)):
            np.set_printoptions(threshold=10000)
            diff = (mask == new_mask).as_numpy_array()
            print(diff.astype(int))
            # print mask.as_numpy_array()
            # print new_mask.as_numpy_array()
            # print (new_mask.as_numpy_array()[:,:,:] %2) * (new_mask.as_numpy_array() == 5)
            assert False


def generate_reflections(detector, beam, scan, experiment, num):
    seed(0)
    assert len(detector) == 1
    beam_vector = flex.vec3_double(num)
    xyzcal_px = flex.vec3_double(num)
    xyzcal_mm = flex.vec3_double(num)
    panel = flex.size_t(num)
    s0_length = matrix.col(beam.get_s0()).length()
    for i in range(num):
        x = randint(0, 2000)
        y = randint(0, 2000)
        z = randint(0, 8)
        s1 = detector[0].get_pixel_lab_coord((x, y))
        s1 = matrix.col(s1).normalize() * s0_length
        phi = scan.get_angle_from_array_index(z, deg=False)
        beam_vector[i] = s1
        xyzcal_px[i] = (x, y, z)
        (x, y) = detector[0].pixel_to_millimeter((x, y))
        xyzcal_mm[i] = (x, y, phi)
        panel[i] = 0

    rlist = flex.reflection_table()
    rlist["id"] = flex.int(len(beam_vector), 0)
    rlist["s1"] = beam_vector
    rlist["panel"] = panel
    rlist["xyzcal.px"] = xyzcal_px
    rlist["xyzcal.mm"] = xyzcal_mm
    rlist["bbox"] = rlist.compute_bbox(experiment)
    index = []
    image_size = experiment[0].detector[0].get_image_size()
    array_range = experiment[0].scan.get_array_range()
    bbox = rlist["bbox"]
    for i in range(len(rlist)):
        x0, x1, y0, y1, z0, z1 = bbox[i]
        if (
            x0 < 0
            or x1 > image_size[0]
            or y0 < 0
            or y1 > image_size[1]
            or z0 < array_range[0]
            or z1 > array_range[1]
        ):
            index.append(i)
    rlist.del_selected(flex.size_t(index))
    rlist["shoebox"] = flex.shoebox(rlist["panel"], rlist["bbox"])
    rlist["shoebox"].allocate_with_value(MaskCode.Valid)
    return rlist



"""
The program dials.damage_analysis calculates dose dependent data quality statistics.

The algorithms and statistics are described in "How best to use photons",
G. Winter et al. 2019 https://doi.org/10.1107/S2059798319003528

Input data can be provided in mtz format or scaled .expt and .refl files.

For dials datafiles, by default, the image number is taken as a proxy for dose.
For multi-sweep datasets, each experiment is assumed to be a measurement of a
new crystal, and is assigned a starting dose of zero. If the multi-sweep
dataset is repeated measurements of one crystal, then the option
shared_crystal=True can be given, which will accumulate dose across the sweeps.
Alternatively, a list of starting_doses can be given. The default assumption is
that each image corresponds to an equivalent dose. This can be overwritten with
the dose_per_image option, where a list of values can be provided to set a
different dose per image for each sweep.

For mtz file input, if there is no 'dose' column, then by default the batch
number is used as a proxy for dose (i.e. dose is accumulated across sweeps),
which may not be suitable for multi-sweep mtz files, unless all sweeps are
measured on the same crystal.

Example usage:

dials.damage_analysis scaled.expt scaled.refl

dials.damage_analysis scaled.mtz

dials.damage_analysis scaled.expt scaled.refl shared_crystal=True

"""

from __future__ import annotations

import json
import logging
import os
import sys

from jinja2 import ChoiceLoader, Environment, PackageLoader

from cctbx import crystal, miller
from iotbx import mtz
from libtbx import phil
from scitbx.array_family import flex

from dials.command_line.symmetry import median_unit_cell
from dials.pychef import Statistics, batches_to_dose, interpret_images_to_doses_options
from dials.util import log, resolution_analysis, show_mail_handle_errors
from dials.util.filter_reflections import filter_reflection_table
from dials.util.options import ArgumentParser, reflections_and_experiments_from_files
from dials.util.version import dials_version

try:
    from typing import List
except ImportError:
    pass

logger = logging.getLogger("dials.command_line.damage_analysis")

phil_scope = phil.parse(
    """\
dose {
    experiments
    {
        dose_per_image = 1
        .type = ints(value_min=0)
        .help = "The 'dose' accumulated per image. If more than one value is"
                "given, this indicates the dose per image for each experiment,"
                "and hence must match the number of experiments."
        starting_doses = None
        .type = ints(value_min=0)
        .help = "The dose values at the start of each sweep. Must match the"
                "number of experiments. If none given, it is assumed that each"
                "sweep starts with zero accumulated dose."
        shared_crystal = False
        .type = bool
        .help = "Option to indicate that all sweeps correspond to measurements"
                "on the same crystal. Therefore the starting doses are"
                "automatically adjusted to account for previously accumulated dose."
    }
}
output {
    log = "dials.damage_analysis.log"
        .type = str
        .help = "The log name"
    html = "dials.damage_analysis.html"
        .type = str
        .help = "Filename for the html report."
    json = None
        .type = str
        .help = "Filename for the html report data in json format."
}
include scope dials.pychef.phil_scope
""",
    process_includes=True,
)


class PychefRunner:

    """Class to prepare input data and run the pychef algorithm."""

    def __init__(self, intensities, dose, params):
        self.params = params
        self.stats = None
        sel = dose >= 0
        self.intensities = intensities.select(sel)
        self.dose = dose.select(sel)
        self.resolution_filter()

    def resolution_filter(self):
        """Filter arrays on resolution."""
        if not self.params.d_min and self.params.min_completeness:
            # Update self.params.d_min using resolution estimate
            params = resolution_analysis.phil_defaults.extract().resolution
            params.nbins = self.params.resolution_bins
            r = resolution_analysis.Resolutionizer(self.intensities, params)
            self.params.d_min = r.resolution(
                resolution_analysis.metrics.COMPLETENESS,
                limit=self.params.min_completeness,
            ).d_min
            logger.info("Estimated d_min: %.2f", self.params.d_min)

        if self.params.d_min or self.params.d_max:
            sel = flex.bool(self.intensities.size(), True)
            d_spacings = self.intensities.d_spacings().data()
            if self.params.d_min:
                sel &= d_spacings >= self.params.d_min
            if self.params.d_max:
                sel &= d_spacings <= self.params.d_max
            self.intensities = self.intensities.select(sel)
            self.dose = self.dose.select(sel)

    @classmethod
    def from_mtz(cls, params, mtz_object):
        """Initialise the class from mtz input.

        Args:
            params: A damage-analysis phil params object
            mtz_object: An iotbx.mtz.object.
        """
        miller_arrays = mtz_object.as_miller_arrays(
            merge_equivalents=False, anomalous=params.anomalous
        )

        intensities = None
        batches = None
        dose = None

        for ma in miller_arrays:
            if ma.info().labels == ["BATCH"]:
                batches = ma
            elif ma.info().labels == ["DOSE"]:
                dose = ma
            elif ma.info().labels == ["I", "SIGI"]:
                intensities = ma
            elif ma.info().labels == ["I(+)", "SIGI(+)", "I(-)", "SIGI(-)"]:
                intensities = ma
        if not intensities:
            raise KeyError("Intensity array not found in mtz file")
        if not batches:
            raise KeyError("Batch array not found in mtz file")

        indices = mtz_object.extract_original_index_miller_indices()
        intensities = intensities.customized_copy(indices=indices)
        batches = batches.customized_copy(indices=indices)

        if params.anomalous:
            intensities = intensities.as_anomalous_array()
            batches = batches.as_anomalous_array()

        if dose is None:
            dose = batches_to_dose(batches.data(), params.dose)
        else:
            dose = dose.data()

        return cls(intensities, dose, params)

    @classmethod
    def from_dials_data_files(cls, params, experiments, reflection_table):
        """Initialise the class from an experiment list and reflection table.

        Args:
            params: A damage-analysis phil params object
            experiments: An ExperimentList
            reflection_table: A reflection table.
        """
        reflection_table = filter_reflection_table(
            reflection_table, intensity_choice=["scale"], partiality_threshold=0.4
        )

        # get scaled intensities
        intensities = miller.array(
            miller.set(
                crystal.symmetry(
                    unit_cell=median_unit_cell(experiments),
                    space_group=experiments[0].crystal.get_space_group(),
                ),
                indices=reflection_table["miller_index"],
                anomalous_flag=params.anomalous,
            ),
            data=reflection_table["intensity.scale.value"],
            sigmas=flex.sqrt(reflection_table["intensity.scale.variance"]),
        )
        intensities.set_observation_type_xray_intensity()

        doses = flex.double()
        start_doses, doses_per_image = interpret_images_to_doses_options(
            experiments,
            params.dose.experiments.dose_per_image,
            params.dose.experiments.starting_doses,
            params.dose.experiments.shared_crystal,
        )
        logger.info(
            "Interpreting data using:\n  starting_doses=%s\n  dose_per_image=%s",
            ", ".join("%s" % i for i in start_doses)
            if len(set(start_doses)) > 1
            else f" all {start_doses[0]}",
            ", ".join("%s" % i for i in doses_per_image)
            if len(set(doses_per_image)) > 1
            else f" all {doses_per_image[0]}",
        )

        for expt, starting_dose, dose_per_img in zip(
            experiments, start_doses, doses_per_image
        ):
            refls = reflection_table.select(expt)
            imgno = flex.ceil(refls["xyzobs.px.value"].parts()[2])
            dose = (imgno * dose_per_img) + starting_dose
            doses.extend(dose)

        doses = doses.iround()

        return cls(intensities, doses, params)

    def run(self):
        """Run the pychef analysis."""
        self.stats = Statistics(
            self.intensities,
            self.dose,
            n_bins=self.params.resolution_bins,
            range_min=self.params.range.min,
            range_max=self.params.range.max,
            range_width=self.params.range.width,
        )

        logger.debug(self.stats.completeness_vs_dose_str())
        logger.debug(self.stats.rcp_vs_dose_str())
        logger.debug(self.stats.scp_vs_dose_str())
        logger.debug(self.stats.rd_vs_dose_str())

    def make_html_report(self, html_filename=None, json_filename=None):
        """Generate html report from pychef stats."""
        data = {"dose_plots": self.stats.to_dict()}
        if html_filename:
            logger.info("Writing html report to: %s", html_filename)
            loader = ChoiceLoader(
                [
                    PackageLoader("dials", "templates"),
                    PackageLoader("dials", "static", encoding="utf-8"),
                ]
            )
            env = Environment(loader=loader)
            template = env.get_template("simple_report.html")
            html = template.render(
                page_title="Damage analysis report",
                panel_title="Damage analysis plots",
                panel_id="dose_plots",
                graphs=data["dose_plots"],
            )
            with open(html_filename, "wb") as f:
                f.write(html.encode("utf-8", "xmlcharrefreplace"))
        if json_filename:
            logger.info("Writing html report data to: %s", json_filename)
            with open(json_filename, "w") as outfile:
                json.dump(data, outfile)


@show_mail_handle_errors()
def run(args: List[str] = None, phil: phil.scope = phil_scope) -> None:
    """Run the command-line script."""

    usage = "dials.damage_analysis [options] scaled.expt scaled.refl | scaled.mtz"

    parser = ArgumentParser(
        usage=usage,
        phil=phil,
        epilog=__doc__,
        read_experiments=True,
        read_reflections=True,
        check_format=False,
    )

    params, _, unhandled = parser.parse_args(
        args=args, show_diff_phil=False, return_unhandled=True
    )

    log.config(logfile=params.output.log)
    logger.info(dials_version())

    reflections, experiments = reflections_and_experiments_from_files(
        params.input.reflections, params.input.experiments
    )
    try:
        if experiments and reflections:
            if len(reflections) != 1:
                raise ValueError("A single input reflections datafile is required")
            if "inverse_scale_factor" not in reflections[0]:
                raise KeyError("Input data must be scaled.")
            script = PychefRunner.from_dials_data_files(
                params,
                experiments,
                reflections[0],
            )

        elif unhandled and os.path.isfile(unhandled[0]):
            try:
                mtz_object = mtz.object(file_name=unhandled[0])
            except RuntimeError:
                # If an error is encountered trying to read the file as an mtzfile
                raise ValueError(
                    "Input file cannot be read as a valid experiment/reflection file or MTZ file"
                )
            else:
                script = PychefRunner.from_mtz(params, mtz_object)
        else:
            parser.print_help()
            raise ValueError("Suitable input datafiles not provided")
    except (ValueError, KeyError) as e:
        sys.exit(f"Error: {e}")
    else:
        script.run()
        script.make_html_report(params.output.html, params.output.json)


if __name__ == "__main__":
    run()



from __future__ import annotations

import logging
import random
import textwrap
from math import log, pi, sqrt
from typing import List

import numpy as np
from numpy.linalg import det, inv, norm

from dxtbx import flumpy
from scitbx import linalg, matrix

from dials.algorithms.profile_model.ellipsoid import mosaicity_from_eigen_decomposition
from dials.algorithms.profile_model.ellipsoid.model import (
    compute_change_of_basis_operation,
)
from dials.algorithms.profile_model.ellipsoid.parameterisation import (
    ReflectionModelState,
)
from dials.array_family import flex
from dials.util import tabulate

logger = logging.getLogger("dials")

flex.set_random_seed(0)
random.seed(0)


class BadSpotForIntegrationException(Exception):
    pass


def compute_dSbar(S: np.array, dS: np.array) -> np.array:
    # dS & S are 3x3 arrays. Returns a 2x2 array
    S12 = S[0:2, 2].reshape(2, 1)
    S21 = S[2, 0:2].reshape(1, 2)
    S22 = S[2, 2]

    dS11 = dS[0:2, 0:2]
    dS12 = dS[0:2, 2].reshape(2, 1)
    dS21 = dS[2, 0:2].reshape(1, 2)
    dS22 = dS[2, 2]

    S22_inv = 1 / S22

    A = dS11
    B = np.matmul(S12 * S22_inv * dS22 * S22_inv, S21)
    C = np.matmul(S12 * S22_inv, dS21)
    D = np.matmul(dS12 * S22_inv, S21)
    return A + B - (C + D)


def compute_dmbar(S: np.array, dS: np.array, dmu: np.array, epsilon: float) -> np.array:
    # S, dS are is a 3x3 array, dmu is 3x1 array
    dmu = dmu.reshape(3, 1)  # 3x1 array

    S12 = S[0:2, 2].reshape(2, 1)
    S22 = S[2, 2]

    dS12 = dS[0:2, 2].reshape(2, 1)
    dS22 = dS[2, 2]

    S22_inv = 1 / S22

    dmu1 = dmu[0:2, 0].reshape(2, 1)
    dmu2 = dmu[2, 0]
    dep = -dmu2

    A = dmu1
    B = dS12 * S22_inv * epsilon
    C = -S12 * S22_inv * dS22 * S22_inv * epsilon
    D = S12 * S22_inv * dep
    return A + B + C + D


class ConditionalDistribution(object):
    """
    A class to compute useful stuff about the conditional distribution

    """

    def __init__(self, norm_s0, mu, dmu, S, dS):
        # norm_s0 is a float i.e. norm(s0)
        self._mu = mu  # 3x1 array
        self._dmu = dmu  # 3 x n array
        self._S = S  # 3x3 array
        self._dS = dS  # 3 x 3 x n array

        # Partition the covariance matrix
        S11 = S[0:2, 0:2]
        S12 = S[0:2, 2].reshape(2, 1)
        S21 = S[2, 0:2].reshape(1, 2)
        S22 = S[2, 2]

        # The partitioned mean vector
        mu1 = mu[0:2, 0].reshape(2, 1)
        mu2 = mu[2, 0]
        # a = norm(s0)

        # The epsilon
        self.epsilon = norm_s0 - mu2

        # Compute the conditional mean
        self._mubar = mu1 + S12 * (1 / S22) * self.epsilon
        assert self._mubar.shape == (2, 1)

        # Compute the conditional covariance matrix
        self._Sbar = S11 - np.matmul(S12 * (1 / S22), S21)
        assert self._Sbar.shape == (2, 2)

        # Set to None and compute on demand
        self.dSbar = None
        self.dmbar = None
        self.d2Sbar = None
        self.d2mbar = None

    def mean(self) -> np.array:
        """
        Return the conditional mean (a 2x1 array)

        """
        return self._mubar

    def sigma(self) -> np.array:
        """
        Return the conditional sigma (a 2x2 array)

        """
        return self._Sbar

    def first_derivatives_of_sigma(self) -> List[np.array]:
        """
        Return the marginal first derivatives (as a list of 2x2 arrays)

        """
        if self.dSbar is None:
            self.dSbar = [
                compute_dSbar(self._S, self._dS[:, :, i])
                for i in range(self._dS.shape[2])
            ]

        return self.dSbar

    def first_derivatives_of_mean(self) -> List[np.array]:
        """
        Return the marginal first derivatives (a list of 2x1 arrays)

        """
        if self.dmbar is None:
            self.dmbar = [
                compute_dmbar(self._S, self._dS[:, :, i], self._dmu[:, i], self.epsilon)
                for i in range(self._dS.shape[2])
            ]

        return self.dmbar


def rotate_vec3_double(R, A):
    """
    Helper function to rotate an array of matrices

    """
    return np.einsum("ij,jk->ik", R, A)


def rotate_mat3_double(R, A):
    """
    Helper function to rotate an array of matrices

    """
    return np.einsum("ij,jkv,kl->ilv", R, A, R.T)


class ReflectionLikelihood(object):
    def __init__(self, model, s0, sp, h, ctot, mobs, sobs):

        # Save stuff
        modelstate = ReflectionModelState(model, s0, h)
        self.modelstate = modelstate
        self.s0 = s0.reshape(3, 1)
        self.norm_s0 = norm(s0)
        self.sp = sp.reshape(3, 1)
        self.h = np.array([h], dtype=np.float64).reshape(3, 1)
        self.ctot = ctot
        self.mobs = mobs.reshape(2, 1)
        self.sobs = sobs

        # Compute the change of basis
        self.R = compute_change_of_basis_operation(self.s0, self.sp)  # const
        s2 = self.s0 + self.modelstate.get_r()
        # Rotate the mean vector
        self.mu = np.matmul(self.R, s2)
        self.S = np.matmul(
            np.matmul(self.R, modelstate.mosaicity_covariance_matrix), self.R.T
        )  # const when not refining mosaicity
        self.dS = rotate_mat3_double(
            self.R, modelstate.get_dS_dp()
        )  # const when not refining mosaicity?
        self.dmu = rotate_vec3_double(
            self.R, modelstate.get_dr_dp()
        )  # const when not refining uc/orientation?
        # Construct the conditional distribution
        self.conditional = ConditionalDistribution(
            self.norm_s0, self.mu, self.dmu, self.S, self.dS
        )

    def update(self):

        # The s2 vector
        s2 = self.s0 + self.modelstate.get_r()
        # Rotate the mean vector
        self.mu = np.matmul(self.R, s2)

        # Rotate the covariance matrix
        if not self.modelstate.state.is_mosaic_spread_fixed:
            self.S = np.matmul(
                np.matmul(self.R, self.modelstate.mosaicity_covariance_matrix), self.R.T
            )  # const when not refining mosaicity

        # Rotate the first derivative matrices
        if not self.modelstate.state.is_mosaic_spread_fixed:
            self.dS = rotate_mat3_double(
                self.R, self.modelstate.get_dS_dp()
            )  # const when not refining mosaicity?

        # Rotate the first derivative of s2
        if (not self.modelstate.state.is_unit_cell_fixed) or not (
            self.modelstate.state.is_orientation_fixed
        ):
            self.dmu = rotate_vec3_double(
                self.R, self.modelstate.get_dr_dp()
            )  # const when not refining uc/orientation?

        # Construct the conditional distribution
        self.conditional = ConditionalDistribution(
            self.norm_s0, self.mu, self.dmu, self.S, self.dS
        )

    def log_likelihood(self):
        """
        Compute the log likelihood for the reflection

        """

        # Get data
        ctot = self.ctot
        mobs = self.mobs
        Sobs = self.sobs

        # Get info about the marginal
        S22 = self.S[2, 2]
        S22_inv = 1 / S22
        mu2 = self.mu[2, 0]

        # Get info about the conditional
        Sbar = self.conditional.sigma()
        mubar = self.conditional.mean()
        Sbar_inv = inv(Sbar)
        Sbar_det = det(Sbar)

        # Weights for marginal and conditional components
        m_w = ctot
        c_w = ctot

        # Compute the marginal likelihood
        m_d = self.norm_s0 - mu2
        m_lnL = m_w * (log(S22) + S22_inv * m_d**2)

        # Compute the conditional likelihood
        c_d = mobs - mubar
        c_lnL = c_w * (
            log(Sbar_det)
            + np.trace(np.matmul(Sbar_inv, (Sobs + np.matmul(c_d, c_d.T))))
        )

        # Return the joint likelihood
        jLL = -0.5 * (m_lnL + c_lnL)
        return jLL

    def first_derivatives(self):
        """
        Compute the first derivatives

        """
        # Get data
        ctot = self.ctot
        mobs = self.mobs  # 2x1 array
        Sobs = self.sobs

        # Get info about marginal distribution
        S22 = self.S[2, 2]
        dS22_vec = self.dS[2, 2, :]  # for i in range(self.dS.shape[2])]
        S22_inv = 1 / S22
        mu2 = self.mu[2, 0]

        # Get info about conditional distribution
        Sbar = self.conditional.sigma()  # 3x3 array
        mubar = self.conditional.mean()  # 2x1 array
        dSbar = self.conditional.first_derivatives_of_sigma()  # list of 2x2 arrays
        dmbar = self.conditional.first_derivatives_of_mean()  # list of 2x1 arrays
        Sbar_inv = inv(Sbar)

        # The distance from the ewald sphere
        epsilon = self.norm_s0 - mu2
        c_d = mobs - mubar  # 2x1 array

        # Weights for marginal and conditional components
        m_w = ctot
        c_w = ctot

        # Compute the derivative wrt parameter i
        I = np.array([[1.0, 0], [0, 1.0]], dtype=np.float64).reshape(2, 2)

        V1 = Sobs + np.matmul(c_d, c_d.T)
        V2 = I - np.matmul(Sbar_inv, V1)

        dSbar = np.array(dSbar)
        dmbar_vec = np.array(dmbar)

        V_vec = np.einsum("ij,ljk->ikl", Sbar_inv, dSbar)
        V_vec = c_w * np.einsum("ijl,ji->l", V_vec, V2)
        dep = -self.dmu[2, :]
        U_vec = m_w * (
            S22_inv * dS22_vec * (1.0 - S22_inv * epsilon**2)
            + 2 * S22_inv * epsilon * dep
        )
        W_vec = np.einsum("ij,lkj->ikl", c_d, dmbar_vec)
        W_vec = -2.0 * c_w * np.einsum("ij,jil->l", Sbar_inv, W_vec)

        dL = -0.5 * (U_vec + V_vec + W_vec)
        return dL

    def fisher_information(self):
        """
        Compute the fisher information

        """
        ctot = self.ctot

        # Get info about marginal distribution
        S22 = self.S[2, 2]
        dS22 = [self.dS[2, 2, i] for i in range(self.dS.shape[2])]
        S22_inv = 1 / S22

        # Get info about conditional distribution
        Sbar = self.conditional.sigma()
        dSbar = self.conditional.first_derivatives_of_sigma()  # list of 2x2 arrays
        dmbar = self.conditional.first_derivatives_of_mean()  # list of 2x1 arrays
        Sbar_inv = inv(Sbar)
        dmu = self.dmu

        # Weights for marginal and conditional components
        m_w = ctot
        c_w = ctot

        # Compute the fisher information wrt parameter i j
        I = flex.double(flex.grid(len(dS22), len(dS22)))

        for j in range(len(dS22)):
            for i in range(len(dS22)):
                U = S22_inv * dS22[j] * S22_inv * dS22[i]
                V = np.trace(
                    np.matmul(
                        np.matmul(
                            np.matmul(
                                Sbar_inv,
                                dSbar[j],
                            ),
                            Sbar_inv,
                        ),
                        dSbar[i],
                    )
                )
                W = 2 * np.trace(
                    np.matmul(
                        np.matmul(
                            Sbar_inv,
                            dmbar[i],
                        ),
                        dmbar[j].T,
                    )
                )
                X = 2 * dmu[2, i] * S22_inv * dmu[2, j]
                I[j, i] = 0.5 * c_w * (V + W) + 0.5 * m_w * (U + X)

        return I


class MaximumLikelihoodTarget(object):
    def __init__(self, model, s0, sp_list, h_list, ctot_list, mobs_list, sobs_list):

        # Check input
        assert len(h_list) == sp_list.shape[-1]
        assert len(h_list) == ctot_list.shape[-1]
        assert len(h_list) == mobs_list.shape[-1]
        assert len(h_list) == sobs_list.shape[-1]

        # Save the model
        self.model = model

        # Compute the change of basis for each reflection
        self.data = []
        for i in range(len(h_list)):
            self.data.append(
                ReflectionLikelihood(
                    model,
                    s0,
                    sp_list[:, i],
                    matrix.col(h_list[i]),
                    ctot_list[i],
                    mobs_list[:, i],
                    sobs_list[:, :, i],
                )
            )

    def update(self):
        for d in self.data:
            d.modelstate.update()  # update the ReflectionModelState
            d.update()  # update the ReflectionLikelihood

    def mse(self):
        """
        The MSE in local reflection coordinates

        """
        mse = 0
        for i in range(len(self.data)):
            mbar = self.data[i].conditional.mean()
            xobs = self.data[i].mobs
            mse += np.dot((xobs - mbar).T, xobs - mbar)
        mse /= len(self.data)
        return mse

    def rmsd(self):
        """
        The RMSD in pixels

        """
        mse = np.array([0.0, 0.0], dtype=np.float64)
        for i in range(len(self.data)):
            R = self.data[i].R
            mbar = self.data[i].conditional.mean()
            xobs = self.data[i].mobs
            norm_s0 = self.data[i].norm_s0

            s1 = np.matmul(
                R.T,
                np.array([mbar[0, 0], mbar[1, 0], norm_s0], dtype=np.float64).reshape(
                    3, 1
                ),
            )
            s3 = np.matmul(
                R.T,
                np.array([xobs[0, 0], xobs[1, 0], norm_s0], dtype=np.float64).reshape(
                    3, 1
                ),
            )
            s1 = matrix.col(flumpy.from_numpy(s1[:, 0]))
            s3 = matrix.col(flumpy.from_numpy(s3[:, 0]))
            xyzcal = self.model.experiment.detector[0].get_ray_intersection_px(s1)
            xyzobs = self.model.experiment.detector[0].get_ray_intersection_px(s3)
            r_x = xyzcal[0] - xyzobs[0]
            r_y = xyzcal[1] - xyzobs[1]
            mse += np.array([r_x**2, r_y**2])
        mse /= len(self.data)
        return np.sqrt(mse)

    def log_likelihood(self):
        """
        The joint log likelihood

        """
        return sum(d.log_likelihood() for d in self.data)

    def jacobian(self):
        """
        Return the Jacobian

        """
        return flex.double([list(d.first_derivatives()) for d in self.data])

    def first_derivatives(self):
        """
        The joint first derivatives

        """
        dL = 0
        for d in self.data:
            dL += d.first_derivatives()
        return dL

    def fisher_information(self):
        """
        The joint fisher information

        """
        return sum(d.fisher_information() for d in self.data)


def line_search(func, x, p, tau=0.5, delta=1.0, tolerance=1e-7):
    """
    Perform a line search
    :param func The function to minimize
    :param x The initial position
    :param p The direction to search
    :param tau: The backtracking parameter
    :param delta: The initial step
    :param tolerance: The algorithm tolerance
    :return: The amount to move in the given direction

    """
    fa = func(x)
    if p.length() < 1:
        min_delta = tolerance
    else:
        min_delta = tolerance / p.length()

    while delta > min_delta:
        try:
            fb = func(x + delta * p)
            if fb <= fa:
                return delta
        except Exception:
            pass
        delta *= tau
    return 0


def gradient_descent(f, df, x0, max_iter=1000, tolerance=1e-10):
    """
    Find the minimum using gradient descent and a line search
    :param f The function to minimize
    :param df The function to compute derivatives
    :param x0 The initial position
    :param max_iter: The maximum number of iterations
    :param tolerance: The algorithm tolerance
    :return: The amount to move in the given direction

    """
    delta = 0.5
    for it in range(max_iter):
        p = -matrix.col(df(x0))
        delta = line_search(f, x0, p, delta=min(1.0, delta * 2), tolerance=tolerance)
        x = x0 + delta * p
        assert f(x) <= f(x0)
        if (x - x0).length() < tolerance:
            break
        x0 = x
    return x


class FisherScoringMaximumLikelihoodBase(object):
    """
    A class to solve maximum likelihood equations using fisher scoring

    """

    def __init__(self, x0, max_iter=1000, tolerance=1e-7):
        """
        Configure the algorithm

        :param x0: The initial parameter estimates
        :param max_iter: The maximum number of iterations
        :param tolerance: The parameter tolerance

        """
        self.x0 = matrix.col(x0)
        self.max_iter = max_iter
        self.tolerance = tolerance

    def solve(self):
        """
        Find the maximum likelihood estimate

        """
        x0 = self.x0

        # Loop through the maximum number of iterations
        for it in range(self.max_iter):
            # Compute the derivative and fisher information at x0
            S, I = self.score_and_fisher_information(x0)

            # Solve the update equation to get direction
            p = matrix.col(self.solve_update_equation(S, I))

            # Perform a line search to ensure that each step results in an increase the
            # in log likelihood. In the rare case where the update does not result in an
            # increase in the likelihood (only observed for absurdly small samples
            # e.g. 2 reflections or when 1 parameter approaches zero) do an iteration
            # of gradient descent
            delta = self.line_search(x0, p)
            if delta > 0:
                x = x0 + delta * p
            else:
                x = self.gradient_search(x0)

            # Call an update
            self.callback(x)
            # Break the loop if the parameters change less than the tolerance
            if (x - x0).length() < self.tolerance:
                break

            # Update the parameter
            x0 = x

        # Save the parameters
        self.num_iter = it + 1
        self.parameters = x

    def solve_update_equation(self, S, I):
        """
        Solve the update equation using cholesky decomposition
        :param S: The score
        :param I: The fisher information
        :return: The parameter delta

        """

        # Construct triangular matrix
        LL = flex.double()
        for j in range(len(S)):
            for i in range(j + 1):
                LL.append(I[j * len(S) + i])

        # Perform the decomposition
        ll = linalg.l_l_transpose_cholesky_decomposition_in_place(LL)
        p = flex.double(S)
        return ll.solve(p)

    def line_search(self, x, p, tau=0.5, delta=1.0, tolerance=1e-7):
        """
        Perform a line search
        :param x The initial position
        :param p The direction to search
        :return: The amount to move in the given direction

        """

        def f(x):
            return -self.log_likelihood(x)

        return line_search(f, x, p, tolerance=self.tolerance)

    def gradient_search(self, x0):
        """
        Find the minimum using gradient descent and a line search
        :param x0 The initial position
        :return: The amount to move in the given direction

        """

        def f(x):
            return -self.log_likelihood(x)

        def df(x):
            return -self.score(x)

        return gradient_descent(f, df, x0, max_iter=1, tolerance=self.tolerance)


class FisherScoringMaximumLikelihood(FisherScoringMaximumLikelihoodBase):
    """
    A class to solve the maximum likelihood equations

    """

    def __init__(
        self,
        model,
        s0,
        sp_list,
        h_list,
        ctot_list,
        mobs_list,
        sobs_list,
        max_iter=1000,
        tolerance=1e-7,
    ):
        """
        Initialise the algorithm:

        """
        # Initialise the super class
        super(FisherScoringMaximumLikelihood, self).__init__(
            model.active_parameters, max_iter=max_iter, tolerance=tolerance
        )

        # Save the parameterisation
        self.model = model

        # Save some stuff
        self.s0 = s0
        self.sp_list = sp_list
        self.h_list = h_list
        self.ctot_list = ctot_list
        self.mobs_list = mobs_list
        self.sobs_list = sobs_list

        # Store the parameter history
        self.history = []

        self._ml_target = MaximumLikelihoodTarget(
            self.model,
            self.s0,
            self.sp_list,
            self.h_list,
            self.ctot_list,
            self.mobs_list,
            self.sobs_list,
        )

        # Print initial
        self.callback(self.model.active_parameters)

    def log_likelihood(self, x):
        """
        :param x: The parameter estimate
        :return: The log likelihood at x

        """
        self.model.active_parameters = x
        self._ml_target.update()
        return self._ml_target.log_likelihood()

    def score(self, x):
        """
        :param x: The parameter estimate
        :return: The score at x

        """
        self.model.active_parameters = x
        self._ml_target.update()
        return flumpy.from_numpy(self._ml_target.first_derivatives())

    def score_and_fisher_information(self, x):
        """
        :param x: The parameter estimate
        :return: The score and fisher information at x

        """
        self.model.active_parameters = x
        self._ml_target.update()
        S = flumpy.from_numpy(self._ml_target.first_derivatives())
        I = self._ml_target.fisher_information()
        return S, I

    def mse(self, x):
        """
        :param x: The parameter estimate
        :return: The MSE at x

        """
        return self._ml_target.mse()

    def rmsd(self, x):
        """
        :param x: The parameter estimate
        :return: The RMSD at x

        """
        return self._ml_target.rmsd()

    def jacobian(self, x):
        """
        :param x: The parameter estimate
        :return: The Jacobian at x

        """
        return self._ml_target.jacobian()

    def condition_number(self, x):
        """
        The condition number of the Jacobian

        """
        from scitbx.linalg.svd import real as svd_real

        svd = svd_real(self.jacobian(x), False, False)
        return max(svd.sigma) / min(svd.sigma)

    def correlation(self, x):
        """
        The correlation of the Jacobian

        """
        J = self.jacobian(x)
        C = flex.double(flex.grid(J.all()[1], J.all()[1]))
        for j in range(C.all()[0]):
            for i in range(C.all()[1]):
                a = J[:, i : i + 1].as_1d()
                b = J[:, j : j + 1].as_1d()
                C[j, i] = flex.linear_correlation(a, b).coefficient()
        return C

    def callback(self, x):
        """
        Handle and update in parameter values

        """
        self.model.active_parameters = x
        self._ml_target.update()
        lnL = self._ml_target.log_likelihood()
        mse = self._ml_target.mse()
        rmsd = self._ml_target.rmsd()

        # Get the unit cell
        unit_cell = self.model.unit_cell.parameters()

        # Get some matrices
        U = self.model.U_matrix.flatten()
        M = self.model.mosaicity_covariance_matrix.flatten()

        # Print some information
        format_string1 = "  Unit cell: (%.3f, %.3f, %.3f, %.3f, %.3f, %.3f)"
        format_string2 = "  | % .6f % .6f % .6f |"
        format_string3 = "  | % .2e % .2e % .2e |"

        logger.info(f"\nIteration: {len(self.history)}")
        if not self.model.is_unit_cell_fixed:
            logger.info("\n" + format_string1 % unit_cell)
        if not self.model.is_orientation_fixed:
            logger.info(
                "\n".join(
                    [
                        "",
                        "  U matrix (orientation)",
                        format_string2 % tuple(U[0:3]),
                        format_string2 % tuple(U[3:6]),
                        format_string2 % tuple(U[6:9]),
                    ]
                )
            )
        if not self.model.is_mosaic_spread_fixed:
            logger.info(
                "\n".join(
                    [
                        "",
                        "  Sigma M",
                        format_string3 % tuple(M[0:3]),
                        format_string3 % tuple(M[3:6]),
                        format_string3 % tuple(M[6:9]),
                    ]
                )
            )

        logger.info(
            "\n".join(
                [
                    "",
                    "  ln(L) = %f" % lnL,
                    "",
                    "  R.M.S.D (local) = %.5g" % sqrt(mse),
                    "",
                    "  R.M.S.D (pixel): X = %.3f, Y = %.3f" % tuple(rmsd),
                ]
            )
        )

        # Append the parameters to the history
        self.history.append(
            {
                "parameters": list(x),
                "likelihood": lnL,
                "unit_cell": unit_cell,
                "orientation": list(U),
                "rlp_mosaicity": list(M),
                "rmsd": tuple(rmsd),
            }
        )


class Refiner(object):
    """
    High level profile refiner class that handles book keeping etc

    """

    def __init__(self, state, data):
        """
        Set the data and initial parameters

        """
        self.s0 = data.s0
        self.h_list = data.h_list
        self.sp_list = data.sp_list
        self.ctot_list = data.ctot_list
        self.mobs_list = data.mobs_list
        self.sobs_list = data.sobs_list
        self.state = state
        self.history = []

    def refine(self):
        """
        Perform the profile refinement

        """
        self.refine_fisher_scoring()

    def refine_fisher_scoring(self):
        """
        Perform the profile refinement

        """

        # Print information
        logger.info("\nComponents to refine:")
        logger.info(" Orientation:       %s" % (not self.state.is_orientation_fixed))
        logger.info(" Unit cell:         %s" % (not self.state.is_unit_cell_fixed))
        logger.info(" RLP mosaicity:     %s" % (not self.state.is_mosaic_spread_fixed))
        logger.info(
            " Wavelength spread: %s\n" % (not self.state.is_wavelength_spread_fixed)
        )

        # Initialise the algorithm
        self.ml = FisherScoringMaximumLikelihood(
            self.state,
            self.s0,
            self.sp_list,
            self.h_list,
            self.ctot_list,
            self.mobs_list,
            self.sobs_list,
        )

        # Solve the maximum likelihood equations
        self.ml.solve()

        # Get the parameters
        self.parameters = flex.double(self.ml.parameters)

        # set the parameters
        self.state.active_parameters = self.parameters

        # Print summary table of refinement.
        rows = []
        headers = ["Iteration", "likelihood", "RMSD (pixel) X,Y"]
        for i, h in enumerate(self.ml.history):
            l = h["likelihood"]
            rmsd = h["rmsd"]
            rows.append([str(i), f"{l:.4f}", f"{rmsd[0]:.3f}, {rmsd[1]:.3f}"])
        logger.info(
            "\nRefinement steps:\n\n" + textwrap.indent(tabulate(rows, headers), " ")
        )

        # Print the eigen values and vectors of sigma_m
        if not self.state.is_mosaic_spread_fixed:
            logger.info("\nDecomposition of Sigma_M:")
            print_eigen_values_and_vectors(
                matrix.sqr(flumpy.from_numpy(self.state.mosaicity_covariance_matrix))
            )

        # Save the history
        self.history = self.ml.history

        # Return the optimizer
        return self.ml

    def correlation(self):
        """
        Return the correlation matrix between parameters

        """
        return self.ml.correlation(self.state.active_parameters)

    def labels(self):
        """
        Return parameter labels

        """
        return self.state.parameter_labels


class RefinerData(object):
    """
    A class for holding the data needed for the profile refinement

    """

    def __init__(self, s0, sp_list, h_list, ctot_list, mobs_list, sobs_list):
        """
        Init the data

        ctot_list is a list of total counts per reflection

        """
        self.s0 = s0
        self.sp_list = sp_list
        self.h_list = h_list
        self.ctot_list = ctot_list
        self.mobs_list = mobs_list
        self.sobs_list = sobs_list

    @classmethod
    def from_reflections(self, experiment, reflections):
        """
        Generate the required data from the reflections

        """

        # Get the beam vector
        s0 = np.array([experiment.beam.get_s0()], dtype=np.float64).reshape(3, 1)

        # Get the reciprocal lattice vector
        h_list = reflections["miller_index"]

        # Initialise the list of observed intensities and covariances
        sp_list = np.zeros(shape=(3, len(h_list)))
        ctot_list = np.zeros(shape=(len(h_list)))
        mobs_list = np.zeros(shape=(2, len(h_list)))
        Sobs_list = np.zeros(shape=(2, 2, len(h_list)))

        logger.info(
            "Computing observed covariance for %d reflections" % len(reflections)
        )
        s0_length = norm(s0)
        assert len(experiment.detector) == 1
        panel = experiment.detector[0]
        sbox = reflections["shoebox"]
        xyzobs = reflections["xyzobs.px.value"]
        for r in range(len(reflections)):

            # The vector to the pixel centroid
            sp = np.array(
                panel.get_pixel_lab_coord(xyzobs[r][0:2]), dtype=np.float64
            ).reshape(3, 1)
            sp *= s0_length / norm(sp)

            # Compute change of basis
            R = compute_change_of_basis_operation(s0, sp)

            # Get data and compute total counts
            data = sbox[r].data
            mask = sbox[r].mask
            bgrd = sbox[r].background

            # Get array of vectors
            i0 = sbox[r].bbox[0]
            j0 = sbox[r].bbox[2]
            assert data.all()[0] == 1
            X = np.zeros(shape=(data.all()[1:]) + (2,), dtype=np.float64)
            C = np.zeros(shape=data.all()[1:], dtype=np.float64)
            ctot = 0
            for j in range(data.all()[1]):
                for i in range(data.all()[2]):
                    c = data[0, j, i] - bgrd[0, j, i]
                    if mask[0, j, i] & (1 | 4) == (1 | 4) and c > 0:
                        ctot += c
                        ii = i + i0
                        jj = j + j0
                        s = np.array(
                            panel.get_pixel_lab_coord((ii + 0.5, jj + 0.5)),
                            dtype=np.float64,
                        ).reshape(3, 1)
                        s *= s0_length / norm(s)
                        e = np.matmul(R, s)
                        X[j, i, :] = e[0:2, 0]
                        C[j, i] = c

            # Check we have a sensible number of counts
            if ctot <= 0:
                raise BadSpotForIntegrationException(
                    "Strong spot found with <= 0 counts! Check spotfinding results"
                )

            # Compute the mean vector
            C = np.expand_dims(C, axis=2)
            xbar = C * X
            xbar = xbar.reshape(-1, xbar.shape[-1]).sum(axis=0)
            xbar /= ctot

            xbar = xbar.reshape(2, 1)

            # Compute the covariance matrix
            Sobs = np.array([[0.0, 0.0], [0.0, 0.0]], dtype=np.float64)
            for j in range(X.shape[0]):
                for i in range(X.shape[1]):
                    x = np.array(X[j, i], dtype=np.float64).reshape(2, 1)
                    Sobs += np.matmul(x - xbar, (x - xbar).T) * C[j, i, 0]

            Sobs /= ctot
            if (Sobs[0, 0] <= 0) or (Sobs[1, 1] <= 0):
                raise BadSpotForIntegrationException(
                    "Strong spot variance <= 0. Check spotfinding results"
                )

            # Add to the lists
            sp_list[:, r] = sp[:, 0]
            ctot_list[r] = ctot
            mobs_list[:, r] = xbar[:, 0]
            Sobs_list[:, :, r] = Sobs

        # Print some information
        logger.info("")
        logger.info(
            "I_min = %.2f, I_max = %.2f" % (np.min(ctot_list), np.max(ctot_list))
        )

        # Sometimes a single reflection might have an enormouse intensity for
        # whatever reason and since we weight by intensity, this can cause the
        # refinement to be dominated by these reflections. Therefore, if the
        # intensity is greater than some value, damp the weighting accordingly
        def damp_outlier_intensity_weights(ctot_list):
            n = ctot_list.size
            sorted_ctot = np.sort(ctot_list)
            Q1 = sorted_ctot[n // 4]
            Q2 = sorted_ctot[n // 2]
            Q3 = sorted_ctot[3 * n // 4]
            IQR = Q3 - Q1
            T = Q3 + 1.5 * IQR
            logger.info(f"Median I = {Q2:.2f}\nQ1/Q3 I = {Q1:.2f}, {Q3:.2f}")
            logger.info(f"Damping effect of intensities > {T:.2f}")
            ndamped = 0
            for i, ctot in enumerate(ctot_list):
                if ctot > T:
                    logger.debug(f"Damping {ctot:.2f}")
                    ctot_list[i] = T
                    ndamped += 1
            logger.info(f"Damped {ndamped}/{n} reflections")
            return ctot_list

        ctot_list = damp_outlier_intensity_weights(ctot_list)

        # Print the mean covariance
        Smean = np.mean(Sobs_list, axis=2)
        logger.info("")
        logger.info("Mean observed covariance:")
        logger.info(print_matrix_np(Smean))
        print_eigen_values_and_vectors_of_observed_covariance(Smean, s0)

        # Compute the distance from the Ewald sphere
        epsilon = flex.double(
            s0_length - matrix.col(s).length() for s in reflections["s2"]
        )
        mv = flex.mean_and_variance(epsilon)
        logger.info("")
        logger.info("Mean distance from Ewald sphere: %.3g" % mv.mean())
        logger.info(
            "Variance in distance from Ewald sphere: %.3g"
            % mv.unweighted_sample_variance()
        )

        # Return the profile refiner data
        return RefinerData(s0, sp_list, h_list, ctot_list, mobs_list, Sobs_list)


def print_eigen_values_and_vectors_of_observed_covariance(A, s0):
    """
    Print the eigen values and vectors of a matrix

    """

    # Compute the eigen decomposition of the covariance matrix
    A = matrix.sqr(flumpy.from_numpy(A))
    s0 = matrix.col(flumpy.from_numpy(s0))
    eigen_decomposition = linalg.eigensystem.real_symmetric(A.as_flex_double_matrix())
    Q = matrix.sqr(eigen_decomposition.vectors())
    L = matrix.diag(eigen_decomposition.values())

    # Print the matrix eigen values
    logger.info(f"\nEigen Values:\n{print_matrix(L, indent=2)}\n")
    logger.info(f"\nEigen Vectors:\n{print_matrix(Q, indent=2)}\n")

    logger.info("Observed covariance in degrees equivalent units")
    logger.info("C1: %.5f degrees" % (sqrt(L[0]) * (180.0 / pi) / s0.length()))
    logger.info("C2: %.5f degrees" % (sqrt(L[3]) * (180.0 / pi) / s0.length()))


def print_eigen_values_and_vectors(A):
    """
    Print the eigen values and vectors of a matrix

    """

    # Compute the eigen decomposition of the covariance matrix
    eigen_decomposition = linalg.eigensystem.real_symmetric(A.as_flex_double_matrix())
    eigen_values = eigen_decomposition.values()

    # Print the matrix eigen values
    logger.info(
        f"\n Eigen Values:\n{print_matrix(matrix.diag(eigen_values), indent=2)}\n"
    )
    logger.info(
        f"\n Eigen Vectors:\n{print_matrix(matrix.sqr(eigen_decomposition.vectors()), indent=2)}\n"
    )

    mosaicity = mosaicity_from_eigen_decomposition(eigen_values)
    logger.info(
        f"""
 Mosaicity in degrees equivalent units:
 M1 : {mosaicity[0]:.5f} degrees
 M2 : {mosaicity[1]:.5f} degrees
 M3 : {mosaicity[2]:.5f} degrees
"""
    )


def print_matrix_np(A, fmt="%.3g", indent=0):
    """
    Pretty print matrix

    """
    t = [fmt % a for a in A.flatten()]
    l = [len(tt) for tt in t]
    max_l = max(l)
    fmt = "%" + ("%d" % (max_l + 1)) + "s"
    prefix = " " * indent
    lines = []
    for j in range(A.shape[0]):
        line = ""
        for i in range(A.shape[1]):
            line += fmt % t[i + j * A.shape[1]]
        lines.append("%s|%s|" % (prefix, line))
    return "\n".join(lines)


def print_matrix(A, fmt="%.3g", indent=0):
    """
    Pretty print matrix

    """
    t = [fmt % a for a in A]
    l = [len(tt) for tt in t]
    max_l = max(l)
    fmt = "%" + ("%d" % (max_l + 1)) + "s"
    prefix = " " * indent
    lines = []
    for j in range(A.n[0]):
        line = ""
        for i in range(A.n[1]):
            line += fmt % t[i + j * A.n[1]]
        lines.append("%s|%s|" % (prefix, line))
    return "\n".join(lines)



from __future__ import annotations

import sys

import pkg_resources

import dials.util

BOLD = "\033[1m"
RED = "\033[1;31m"
GREEN = "\033[32m"
NC = "\033[0m"


def read_entry_point(entry_point):
    """
    Finds all registered plugins for a given entry point.

    :return: A dictionary of entry point plugins
    """
    return {e.name: e for e in pkg_resources.iter_entry_points(entry_point)}


def installation_is_valid():
    """
    Verifies that all required plugins are present.

    :return: True if so, False otherwise
    """
    for ep, ep_dict in known_entry_points.items():
        required_plugins = set(ep_dict.get("required", []))
        if not required_plugins:
            continue
        plugins = read_entry_point(ep)
        missing_plugins = required_plugins - set(plugins)
        if missing_plugins:
            return False
    return True


known_entry_points = {
    "dxtbx.profile_model": {
        "description": "profile models",
        "required": ["gaussian_rs"],
    },
    "dxtbx.scaling_model_ext": {"description": "scaling models"},
    "dials.index.basis_vector_search": {
        "description": "Basis vector search strategies",
        "required": ["fft1d", "fft3d", "real_space_grid_search"],
    },
    "dials.index.lattice_search": {
        "description": "Lattice search strategies",
        "required": ["low_res_spot_match"],
    },
    "dials.spotfinder.threshold": {
        "description": "Spotfinding threshold algorithms",
        "required": ["dispersion", "dispersion_extended", "radial_profile"],
    },
}


@dials.util.show_mail_handle_errors()
def run(_=None):
    for ep, ep_dict in known_entry_points.items():
        print(f"{BOLD}{ep}{NC}  {ep_dict['description']}")
        plugins = read_entry_point(ep)
        for p in sorted(plugins):
            print(
                " {GREEN}{p} {NC}({pp.module_name} via {BOLD}{pp.dist.project_name}{NC} {pp.dist.version})".format(
                    BOLD=BOLD, p=p, GREEN=GREEN, NC=NC, pp=plugins[p]
                )
            )
        required_plugins = set(ep_dict.get("required", []))
        missing_plugins = required_plugins - set(plugins)
        if missing_plugins:
            print(f" {RED}  --- missing required plugins:{NC}")
            for p in missing_plugins:
                print(f" {RED}{p}{NC}")
        print()
    sys.exit(not installation_is_valid())


if __name__ == "__main__":
    run()



from __future__ import annotations

import procrunner
import pytest


@pytest.mark.parametrize("dataset", ["centroid_test_data", "thaumatin_grid_scan"])
def test_spot_counts_per_image(dataset, dials_data, tmp_path):
    path = dials_data(dataset, pathlib=True)

    # import the data
    result = procrunner.run(
        ["dials.import", "output.experiments=imported.expt"]
        + list(path.glob("*.cbf*")),
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    assert tmp_path.joinpath("imported.expt").is_file()

    # find the spots
    result = procrunner.run(
        ["dials.find_spots", "imported.expt", "nproc=1", "min_spot_size=3"],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    assert tmp_path.joinpath("strong.refl").is_file()

    result = procrunner.run(
        [
            "dials.spot_counts_per_image",
            "imported.expt",
            "strong.refl",
            "plot=spot_counts.png",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    assert tmp_path.joinpath("spot_counts.png").is_file()

    assert all(
        s in result.stdout
        for s in (
            b"image",
            b"#spots_no_ice",
            b"total_intensity",
            b"d_min (distl method 1)",
            b"d_min (distl method 2)",
        )
    ), result.stdout


def test_spot_counts_per_image_fails_cleanly_on_indexed(dials_data, tmp_path):
    path = dials_data("insulin_processed", pathlib=True)
    result = procrunner.run(
        [
            "dials.spot_counts_per_image",
            path / "indexed.expt",
            path / "indexed.refl",
        ],
        working_directory=tmp_path,
    )
    assert result.returncode
    assert b"Only unindexed reflections are currently supported" in result.stderr



"""
Tests for the active parameter manager module.
"""

from __future__ import annotations

from unittest.mock import Mock

import pytest

from scitbx import sparse

from dials.algorithms.scaling.active_parameter_managers import (
    ParameterManagerGenerator,
    active_parameter_manager,
    multi_active_parameter_manager,
    shared_active_parameter_manager,
)
from dials.algorithms.scaling.parameter_handler import (
    ScalingParameterManagerGenerator,
    scaling_active_parameter_manager,
)
from dials.algorithms.scaling.target_function import ScalingTarget
from dials.array_family import flex


def mock_component():
    """Return a mock component of a general model."""
    component = Mock()
    component.free_parameters = flex.double([1.0])
    component.free_parameter_esds = None
    component.n_params = 1
    component.var_cov_matrix = sparse.matrix(1, 1)
    return component


def mock_scaling_component(n_refl):
    """Return a mock component of a general model."""
    component = mock_component()
    component.calculate_scales.return_value = flex.double(n_refl, 1.0)
    component.n_refl = [n_refl]
    return component


def mock_data_manager(components):
    """Return a mock data manager of a general model."""
    dm = Mock()
    dm.components = components
    dm.fixed_components = []
    return dm


def test_general_apm():
    """Test for the general active_parameter_manage class."""
    components = {
        "scale": mock_component(),
        "decay": mock_component(),
        "absorption": mock_component(),
    }

    apm = active_parameter_manager(components, ["scale", "decay"])
    assert "decay" in apm.components_list
    assert "scale" in apm.components_list
    assert "absorption" not in apm.components_list
    assert apm.n_active_params == (
        components["scale"].n_params + components["decay"].n_params
    )
    n_cumul = 0
    for component in apm.components:
        assert apm.components[component]["n_params"] == components[component].n_params
        assert apm.components[component]["start_idx"] == n_cumul
        assert (
            apm.components[component]["end_idx"]
            == n_cumul + apm.components[component]["n_params"]
        )
        n_cumul += apm.components[component]["n_params"]

    apm.set_param_vals(flex.double([2.0, 1.5]))
    assert apm.get_param_vals() == flex.double([2.0, 1.5])
    # Test params were updated in components
    assert list(components["scale"].free_parameters) == [2.0]
    assert list(components["decay"].free_parameters) == [1.5]
    # Test selection of parameters
    decay_params = apm.select_parameters("decay")
    assert len(decay_params) == 1
    assert decay_params[0] == 1.5

    # Test calculate model state uncertainties
    var_cov = flex.double([1.0, 0.5, 0.5, 2.0])
    var_cov.reshape(flex.grid(2, 2))
    apm.calculate_model_state_uncertainties(var_cov)
    assert components["scale"].var_cov_matrix[0, 0] == 1.0
    assert components["decay"].var_cov_matrix[0, 0] == 2.0

    # Test set param esds.
    apm.set_param_esds(flex.double([0.1, 0.2]))
    assert components["scale"].free_parameter_esds == flex.double([0.1])
    assert components["decay"].free_parameter_esds == flex.double([0.2])


def test_multi_apm():
    """Test for the general multi_active_parameter_manage class."""

    components_1 = {
        "scale": mock_component(),
        "decay": mock_component(),
        "absorption": mock_component(),
    }
    components_2 = {"scale": mock_component(), "decay": mock_component()}

    multi_apm = multi_active_parameter_manager(
        ScalingTarget(),
        [components_1, components_2],
        [["scale", "decay"], ["scale"]],
        active_parameter_manager,
    )

    # Test correct setup of apm_list attribute.
    for apm in multi_apm.apm_list:
        assert isinstance(apm, active_parameter_manager)
    assert len(multi_apm.apm_list) == 2
    assert multi_apm.components_list == ["scale", "decay", "scale"]
    assert multi_apm.n_active_params == 3
    assert multi_apm.apm_data[0] == {"start_idx": 0, "end_idx": 2}
    assert multi_apm.apm_data[1] == {"start_idx": 2, "end_idx": 3}

    # Test parameter selection.
    multi_apm.set_param_vals(flex.double([3.0, 2.5, 2.0]))
    assert multi_apm.get_param_vals() == flex.double([3.0, 2.5, 2.0])
    assert multi_apm.select_parameters(0) == flex.double([3.0, 2.5])
    assert multi_apm.select_parameters(1) == flex.double([2.0])

    # Test setting parameter esds.
    multi_apm.set_param_esds(flex.double([0.1, 0.2, 0.3]))
    assert components_1["scale"].free_parameter_esds == flex.double([0.1])
    assert components_1["decay"].free_parameter_esds == flex.double([0.2])
    assert components_2["scale"].free_parameter_esds == flex.double([0.3])

    # Test setting var_cov matrices for each component.
    var_cov = flex.double([1.0, 0.5, 0.5, 0.5, 2.0, 0.5, 0.5, 0.5, 3.0])
    var_cov.reshape(flex.grid(3, 3))
    multi_apm.calculate_model_state_uncertainties(var_cov)
    assert components_1["scale"].var_cov_matrix[0, 0] == 1.0
    assert components_1["decay"].var_cov_matrix[0, 0] == 2.0
    assert components_2["scale"].var_cov_matrix[0, 0] == 3.0


def test_shared_apm():
    components_1 = {
        "scale": mock_component(),
        "decay": mock_component(),
        "absorption": mock_component(),
    }
    components_2 = {"scale": mock_component(), "decay": mock_component()}

    multi_apm = shared_active_parameter_manager(
        ScalingTarget(),
        [components_1, components_2],
        [["scale", "decay"], ["scale", "decay"]],
        active_parameter_manager,
        shared="decay",
    )

    # Test correct setup of apm_list attribute.
    for apm in multi_apm.apm_list:
        assert isinstance(apm, active_parameter_manager)
    assert len(multi_apm.apm_list) == 2
    assert multi_apm.components_list == ["scale", "decay", "scale", "decay"]
    assert multi_apm.n_active_params == 4
    # assert multi_apm.apm_data[0] == {"start_idx": 0, "end_idx": 2}
    # assert multi_apm.apm_data[1] == {"start_idx": 2, "end_idx": 4}
    assert list(multi_apm.reducing_matrix.as_dense_matrix()) == [
        1,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        1,
        0,
        1,
        0,
    ]
    assert multi_apm.apm_data[0]["start_idx"] == 0
    assert multi_apm.apm_data[0]["end_idx"] == 2
    assert multi_apm.apm_data[1]["start_idx"] == 2
    assert multi_apm.apm_data[1]["end_idx"] == 4
    assert list(multi_apm.apm_data[0]["apm_sel"]) == [0, 1]
    assert list(multi_apm.apm_data[1]["apm_sel"]) == [2, 1]


def test_ParameterManagerGenerator_concurrent():
    """Test the apm factory for concurrent refinement."""
    components_1 = {
        "scale": mock_component(),
        "decay": mock_component(),
        "absorption": mock_component(),
    }
    data_manager = mock_data_manager(components_1)

    pmg = ParameterManagerGenerator(
        [data_manager],
        apm_type=active_parameter_manager,
        target=ScalingTarget(),
        mode="concurrent",
    )
    apms = pmg.parameter_managers()
    assert len(apms) == 1
    apm = apms[0]
    assert isinstance(apm, multi_active_parameter_manager)
    assert "scale" in apm.components_list
    assert "decay" in apm.components_list
    assert "absorption" in apm.components_list

    components_1 = {
        "scale": mock_component(),
        "decay": mock_component(),
        "absorption": mock_component(),
    }
    components_2 = {"1": mock_component(), "2": mock_component()}
    data_manager_1 = mock_data_manager(components_1)
    data_manager_2 = mock_data_manager(components_2)

    pmg = ParameterManagerGenerator(
        [data_manager_1, data_manager_2],
        apm_type=active_parameter_manager,
        target=ScalingTarget(),
        mode="concurrent",
    )
    multi_apms = pmg.parameter_managers()
    assert len(multi_apms) == 1
    multi_apm = multi_apms[0]
    assert isinstance(multi_apm, multi_active_parameter_manager)
    for apm in multi_apm.apm_list:
        assert isinstance(apm, active_parameter_manager)
    assert "scale" in multi_apm.apm_list[0].components_list
    assert "decay" in multi_apm.apm_list[0].components_list
    assert "absorption" in multi_apm.apm_list[0].components_list
    assert "1" in multi_apm.apm_list[1].components_list
    assert "2" in multi_apm.apm_list[1].components_list

    # now try fixing a component
    data_manager.fixed_components = ["absorption"]
    pmg = ParameterManagerGenerator(
        [data_manager],
        apm_type=active_parameter_manager,
        target=ScalingTarget(),
        mode="concurrent",
    )
    apms = pmg.parameter_managers()
    assert len(apms) == 1
    apm = apms[0]
    assert isinstance(apm, multi_active_parameter_manager)
    assert "scale" in apm.components_list
    assert "decay" in apm.components_list
    assert "absorption" not in apm.components_list


def test_ParameterManagerGenerator_consecutive():
    """Test the apm factory for consecutive refinement."""
    components_1 = {
        "scale": mock_component(),
        "decay": mock_component(),
        "absorption": mock_component(),
    }

    data_manager = mock_data_manager(components_1)
    data_manager.consecutive_refinement_order = [["scale", "decay"], ["absorption"]]

    # Test single dataset case.
    pmg = ParameterManagerGenerator(
        [data_manager],
        apm_type=active_parameter_manager,
        target=ScalingTarget(),
        mode="consecutive",
    )
    apms = list(pmg.parameter_managers())
    assert len(apms) == 2
    apm = apms[0]
    assert isinstance(apm, multi_active_parameter_manager)
    assert "scale" in apm.components_list
    assert "decay" in apm.components_list
    assert "absorption" not in apm.components_list
    apm = apms[1]
    assert isinstance(apm, multi_active_parameter_manager)
    assert "scale" not in apm.components_list
    assert "decay" not in apm.components_list
    assert "absorption" in apm.components_list

    # Test multi dataset case.
    components_2 = {"1": mock_component(), "2": mock_component()}
    data_manager_2 = mock_data_manager(components_2)
    data_manager_2.consecutive_refinement_order = [["1"], ["2"]]

    pmg = ParameterManagerGenerator(
        [data_manager, data_manager_2],
        apm_type=active_parameter_manager,
        target=ScalingTarget(),
        mode="consecutive",
    )
    apms = list(pmg.parameter_managers())
    assert len(apms) == 2
    multi_apm = apms[0]
    assert isinstance(multi_apm, multi_active_parameter_manager)
    apm_1 = multi_apm.apm_list[0]
    assert "scale" in apm_1.components_list
    assert "decay" in apm_1.components_list
    assert "absorption" not in apm_1.components_list
    assert multi_apm.apm_list[1].components_list == ["1"]
    multi_apm = apms[1]
    assert isinstance(multi_apm, multi_active_parameter_manager)
    assert multi_apm.apm_list[0].components_list == ["absorption"]
    assert multi_apm.apm_list[1].components_list == ["2"]

    # Test multi dataset case with different number of cycles for each data_manager.
    components_2 = {"1": mock_component()}
    data_manager_2 = mock_data_manager(components_2)
    data_manager_2.consecutive_refinement_order = [["1"], ["2"]]
    pmg = ParameterManagerGenerator(
        [data_manager, data_manager_2],
        apm_type=active_parameter_manager,
        target=ScalingTarget(),
        mode="consecutive",
    )
    assert pmg.param_lists[0] == [["scale", "decay"], ["absorption"]]
    assert pmg.param_lists[1] == [["1"]]
    apms = list(pmg.parameter_managers())
    assert len(apms) == 2
    multi_apm = apms[0]
    assert isinstance(multi_apm, multi_active_parameter_manager)
    apm_1 = multi_apm.apm_list[0]
    assert "scale" in apm_1.components_list
    assert "decay" in apm_1.components_list
    assert "absorption" not in apm_1.components_list
    assert multi_apm.apm_list[1].components_list == ["1"]
    multi_apm = apms[1]
    assert isinstance(multi_apm, multi_active_parameter_manager)
    assert multi_apm.apm_list[0].components_list == ["absorption"]
    # Only change relative to previous test case.
    assert multi_apm.apm_list[1].components_list == []

    # Test fixing the decay parameter.
    data_manager.fixed_components = ["decay"]
    pmg = ParameterManagerGenerator(
        [data_manager],
        apm_type=active_parameter_manager,
        target=ScalingTarget(),
        mode="consecutive",
    )
    apms = list(pmg.parameter_managers())
    assert len(apms) == 2
    apm = apms[0]
    assert isinstance(apm, multi_active_parameter_manager)
    assert "scale" in apm.components_list
    assert "decay" not in apm.components_list
    assert "absorption" not in apm.components_list
    apm = apms[1]
    assert isinstance(apm, multi_active_parameter_manager)
    assert "scale" not in apm.components_list
    assert "decay" not in apm.components_list
    assert "absorption" in apm.components_list


def test_scaling_active_parameter_manager():
    """Test the scaling-specific parameter manager."""
    components_2 = {"1": mock_scaling_component(2), "2": mock_scaling_component(2)}
    scaling_apm = scaling_active_parameter_manager(components_2, ["1"])
    assert list(scaling_apm.constant_g_values[0]) == list(
        components_2["2"].calculate_scales()
    )
    assert len(scaling_apm.constant_g_values) == 1
    assert scaling_apm.n_obs == [2]

    # Test that no constant_g_values if both components selected
    scaling_apm = scaling_active_parameter_manager(components_2, ["1", "2"])
    assert scaling_apm.constant_g_values is None

    # Check that one can't initialise with an unequal number of reflections,
    # either within the selection or overall.
    with pytest.raises(AssertionError):
        components_2 = {"1": mock_scaling_component(2), "2": mock_scaling_component(1)}
        scaling_apm = scaling_active_parameter_manager(components_2, ["1", "2"])
    with pytest.raises(AssertionError):
        components_2 = {"1": mock_scaling_component(2), "2": mock_scaling_component(1)}
        scaling_apm = scaling_active_parameter_manager(components_2, ["1"])

    data_manager = mock_data_manager(components_2)
    pmg = ScalingParameterManagerGenerator(
        [data_manager], target=ScalingTarget(), mode="concurrent"
    )
    assert isinstance(pmg.apm_type, type(scaling_active_parameter_manager))



"""
Various tools/controls used by the image viewer
"""


from __future__ import annotations

import wx
import wx.lib.newevent
from orderedset import OrderedSet

ZeroMQEvent, EVT_ZEROMQ_EVENT = wx.lib.newevent.NewEvent()


class ImageCollectionWithSelection(OrderedSet):
    """A Ordered-like object that tracks a 'currently selected item'"""

    def __init__(self, items=None):
        super().__init__(items or [])
        self._selected_index = None

    @property
    def selected_index(self):
        return self._selected_index

    @selected_index.setter
    def selected_index(self, index):
        if index is not None:
            index = int(index)
            if index < 0 or index > len(self._items) - 1:
                raise IndexError("Out of index range")
        self._selected_index = int(index)

    @property
    def selected(self):
        if self._selected_index is None:
            return None
        return self[self._selected_index]

    @selected.setter
    def selected(self, item):
        self._selected_index = self.index(item)


class LegacyChooserAdapter:
    """Fake wx.Choice replacement for legacy upstream code.

    The design relies on a wx.Choice object being created by a subclass and
    stored on a frame.image_chooser attribute. This wx.Choice object is then
    used to control the current selected image and storing information about
    all the 'loaded' images. This class replaces the API minimally so that
    the upstream code still works, while letting us transition to something
    a little saner in our implementation.
    """

    def __init__(self, images, loader):
        """
        Create the fake wx.Choice object.
        :param images: The list-like object containing the image data
        :param loader: A function to call to load a specific entry as current
        """
        super().__init__()
        self._images = images
        self._loader = loader

    def GetCount(self):
        return len(self._images)

    def SetSelection(self, index):
        if self._images.selected_index != index:
            # self._images.selected_index = index
            # print("Posting event from fake choice")
            # wx.PostEvent(self, wx.PyCommandEvent(wx.EVT_CHOICE.typeId, 1))
            self._loader(self._images[index])

    def GetSelection(self):
        return self._images.selected_index

    def GetClientData(self, index):
        return self._images[index]


class ImageChooserControl(wx.Control):
    """
    Convenience control to display a slider and accompanying label
    """

    def __init__(self, *args, **kwargs):
        kwargs["style"] = kwargs.get("style", 0) | wx.BORDER_NONE
        super().__init__(*args, **kwargs)

        self._slider = wx.Slider(self, -1, style=wx.SL_AUTOTICKS)
        self._slider.SetMin(1)
        self._label = wx.StaticText(self, -1, "Some Text")

        # Work out the maximum size of the text so that we can cut off the slider to allow room
        _, size_y = self._label.GetEffectiveMinSize()
        self._label.SetFont(self._label.GetFont().Italic())
        self.size_y = max(size_y, self._label.GetEffectiveMinSize()[1])

        # Use a horizontal box to control vertical alignment
        labelSizer = wx.BoxSizer(wx.HORIZONTAL)
        labelSizer.Add(self._label, flag=wx.ALL | wx.ALIGN_BOTTOM)

        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(self._slider, flag=wx.ALL | wx.ALIGN_LEFT)
        sizer.Add(labelSizer, proportion=1, flag=wx.EXPAND)
        self.SetSizer(sizer)

    def Layout(self):
        # Fix the slider height so that we can fit the whole text in
        w, h = self.GetSize()
        self._slider.SetMinSize((w, h - self.size_y))
        # Delegate to the sizers for the rest of the calculation
        super().Layout()

    def set_temporary_label(self, label):
        """Set the display label to a 'temporary' styled entry"""
        self._label.SetForegroundColour((100, 100, 100))
        self._label.SetLabel(label)
        self._label.SetFont(wx.NORMAL_FONT.Italic())
        # Don't seem to get automatic layout calls inside toolbars
        self.Layout()

    def set_label(self, label):
        """Set the display label to a 'permanent' styled entry"""
        self._label.SetForegroundColour((0, 0, 0))
        self._label.SetLabel(label)
        self._label.SetFont(wx.NORMAL_FONT)
        # Don't seem to get automatic layout calls inside toolbars
        self.Layout()

    def GetValue(self):
        return self._slider.GetValue()

    def SetValue(self, value):
        self._slider.SetValue(value)

    def SetMax(self, value):
        "Set maximum, and disable the slider if necessary"
        if value == 1:
            assert self._slider.GetValue() == 1
            self._slider.Disable()
        else:
            self._slider.Enable()
        self._slider.SetMax(value)



"""
A simple test of stills refinement using fake data.
Only the crystal is perturbed while the beam and detector are known.
"""


from __future__ import annotations


def test(args=[]):
    # Python and cctbx imports
    from math import pi

    from cctbx.sgtbx import space_group, space_group_symbols

    # Symmetry constrained parameterisation for the unit cell
    from cctbx.uctbx import unit_cell

    # We will set up a mock scan and a mock experiment list
    from dxtbx.model import ScanFactory
    from dxtbx.model.experiment_list import Experiment, ExperimentList
    from libtbx.phil import parse
    from libtbx.test_utils import approx_equal
    from rstbx.symmetry.constraints.parameter_reduction import symmetrize_reduce_enlarge
    from scitbx import matrix

    # Crystal parameterisations
    from dials.algorithms.refinement.parameterisation.crystal_parameters import (
        CrystalOrientationParameterisation,
        CrystalUnitCellParameterisation,
    )
    from dials.algorithms.refinement.prediction.managed_predictors import (
        ScansRayPredictor,
        StillsExperimentsPredictor,
    )

    # Reflection prediction
    from dials.algorithms.spot_prediction import IndexGenerator, ray_intersection

    # Import for surgery on reflection_tables
    from dials.array_family import flex

    # Get module to build models using PHIL
    from . import geometry_phil, minimiser_phil, setup_geometry

    #############################
    # Setup experimental models #
    #############################

    master_phil = parse(f"{geometry_phil}\n{minimiser_phil}")

    # build models, with a larger crystal than default in order to get enough
    # reflections on the 'still' image
    param = """
  geometry.parameters.crystal.a.length.range=40 50;
  geometry.parameters.crystal.b.length.range=40 50;
  geometry.parameters.crystal.c.length.range=40 50;
  geometry.parameters.random_seed = 42"""
    models = setup_geometry.Extract(
        master_phil, cmdline_args=args, local_overrides=param
    )

    crystal = models.crystal
    mydetector = models.detector
    mygonio = models.goniometer
    mybeam = models.beam

    # Build a mock scan for a 1.5 degree wedge. Only used for generating indices near
    # the Ewald sphere
    sf = ScanFactory()
    myscan = sf.make_scan(
        image_range=(1, 1),
        exposure_times=0.1,
        oscillation=(0, 1.5),
        epochs=list(range(1)),
        deg=True,
    )
    sequence_range = myscan.get_oscillation_range(deg=False)
    im_width = myscan.get_oscillation(deg=False)[1]
    assert approx_equal(im_width, 1.5 * pi / 180.0)

    # Build experiment lists
    stills_experiments = ExperimentList()
    stills_experiments.append(
        Experiment(beam=mybeam, detector=mydetector, crystal=crystal, imageset=None)
    )
    scans_experiments = ExperimentList()
    scans_experiments.append(
        Experiment(
            beam=mybeam,
            detector=mydetector,
            crystal=crystal,
            goniometer=mygonio,
            scan=myscan,
            imageset=None,
        )
    )

    ##########################################################
    # Parameterise the models (only for perturbing geometry) #
    ##########################################################

    xlo_param = CrystalOrientationParameterisation(crystal)
    xluc_param = CrystalUnitCellParameterisation(crystal)

    ################################
    # Apply known parameter shifts #
    ################################

    # rotate crystal (=5 mrad each rotation)
    xlo_p_vals = []
    p_vals = xlo_param.get_param_vals()
    xlo_p_vals.append(p_vals)
    new_p_vals = [a + b for a, b in zip(p_vals, [5.0, 5.0, 5.0])]
    xlo_param.set_param_vals(new_p_vals)

    # change unit cell (=1.0 Angstrom length upsets, 0.5 degree of
    # gamma angle)
    xluc_p_vals = []
    p_vals = xluc_param.get_param_vals()
    xluc_p_vals.append(p_vals)
    cell_params = crystal.get_unit_cell().parameters()
    cell_params = [a + b for a, b in zip(cell_params, [1.0, 1.0, -1.0, 0.0, 0.0, 0.5])]
    new_uc = unit_cell(cell_params)
    newB = matrix.sqr(new_uc.fractionalization_matrix()).transpose()
    S = symmetrize_reduce_enlarge(crystal.get_space_group())
    S.set_orientation(orientation=newB)
    X = tuple([e * 1.0e5 for e in S.forward_independent_parameters()])
    xluc_param.set_param_vals(X)

    # keep track of the target crystal model to compare with refined
    from copy import deepcopy

    target_crystal = deepcopy(crystal)

    #############################
    # Generate some reflections #
    #############################

    # All indices in a 2.0 Angstrom sphere for crystal
    resolution = 2.0
    index_generator = IndexGenerator(
        crystal.get_unit_cell(),
        space_group(space_group_symbols(1).hall()).type(),
        resolution,
    )
    indices = index_generator.to_array()

    # Build a ray predictor and predict rays close to the Ewald sphere by using
    # the narrow rotation scan
    ref_predictor = ScansRayPredictor(scans_experiments, sequence_range)
    obs_refs = ref_predictor(indices, experiment_id=0)

    # Take only those rays that intersect the detector
    intersects = ray_intersection(mydetector, obs_refs)
    obs_refs = obs_refs.select(intersects)

    # Add in flags and ID columns by copying into standard reflection table
    tmp = flex.reflection_table.empty_standard(len(obs_refs))
    tmp.update(obs_refs)
    obs_refs = tmp

    # Invent some variances for the centroid positions of the simulated data
    im_width = 0.1 * pi / 180.0
    px_size = mydetector[0].get_pixel_size()
    var_x = flex.double(len(obs_refs), (px_size[0] / 2.0) ** 2)
    var_y = flex.double(len(obs_refs), (px_size[1] / 2.0) ** 2)
    var_phi = flex.double(len(obs_refs), (im_width / 2.0) ** 2)
    obs_refs["xyzobs.mm.variance"] = flex.vec3_double(var_x, var_y, var_phi)

    # Re-predict using the stills reflection predictor
    stills_ref_predictor = StillsExperimentsPredictor(stills_experiments)
    obs_refs_stills = stills_ref_predictor(obs_refs)

    # Set 'observed' centroids from the predicted ones
    obs_refs_stills["xyzobs.mm.value"] = obs_refs_stills["xyzcal.mm"]

    ###############################
    # Undo known parameter shifts #
    ###############################

    xlo_param.set_param_vals(xlo_p_vals[0])
    xluc_param.set_param_vals(xluc_p_vals[0])

    # make a refiner
    from dials.algorithms.refinement.refiner import phil_scope

    params = phil_scope.fetch(source=parse("")).extract()

    # Change this to get a plot
    do_plot = False
    if do_plot:
        params.refinement.refinery.journal.track_parameter_correlation = True

    from dials.algorithms.refinement.refiner import RefinerFactory

    # decrease bin_size_fraction to terminate on RMSD convergence
    params.refinement.target.bin_size_fraction = 0.01
    params.refinement.parameterisation.beam.fix = "all"
    params.refinement.parameterisation.detector.fix = "all"
    refiner = RefinerFactory.from_parameters_data_experiments(
        params, obs_refs_stills, stills_experiments
    )

    # run refinement
    history = refiner.run()

    # regression tests
    assert len(history["rmsd"]) == 9

    refined_crystal = refiner.get_experiments()[0].crystal
    uc1 = refined_crystal.get_unit_cell()
    uc2 = target_crystal.get_unit_cell()
    assert uc1.is_similar_to(uc2)

    if do_plot:
        plt = refiner.parameter_correlation_plot(
            len(history["parameter_correlation"]) - 1
        )
        plt.show()
